"""
This subpackage contains procedures / routines built on top of the main package classes.

"""

__all__ = ["psf_routines", "deconvolution_routines"]

from .psf_routines import build_psf, run_multi_steps_PSF_reconstruction
from .deconvolution_routines import multi_steps_deconvolution



from time import time

import jax.numpy as jnp
import jax.scipy.signal
import numpy as np
from astropy.io import fits
from jax import lax, jit
from jax.numpy.fft import fft2, ifft2
from functools import partial
from copy import deepcopy


def timer_func(func):
    """
    This function shows the execution time of the provided function object.

    """

    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f'Function {func.__name__!r} executed in {(t2 - t1):.4f}s')
        return result

    return wrap_func

@partial(jit, static_argnums=(2,))
def pad_and_convolve(x, y, padding=True):
    """
    Lax convolution is performed after an optional padding.

    :param x: first array 
    :param y: second array 
    :param padding: padding
    :type padding: bool

    :return: output array

    """
    x = x.reshape(1, x.shape[0], x.shape[1], 1)
    y = y.reshape(y.shape[0], y.shape[1], 1, 1)

    if padding:
        x = jnp.pad(x, ((0, 0), (4, 4), (4, 4), (0, 0)), mode='wrap')

    dimension_numbers = ('NHWC', 'HWIO', 'NHWC')
    dn = lax.conv_dimension_numbers(x.shape, y.shape, dimension_numbers)

    output = lax.conv_general_dilated(x, y, (1, 1), 'SAME', (1, 1), (1, 1), dn)

    if padding:
        return output[:, 4:-4, 4:-4, :]
    else:
        return output

@partial(jit, static_argnums=(2,))
def pad_and_convolve_fft(x, y, padding=True):
    """
    FFT (Fast Fourier Transform) convolution is performed after an optional padding.

    :param x: first array 
    :param y: second array 
    :param padding: padding
    :type padding: bool

    :return: Output array

    """
    if padding:
        x = jnp.pad(x, ((4, 4), (4, 4)), constant_values=0.)
        y = jnp.pad(y, ((4, 4), (4, 4)), constant_values=0.)

    output = fft_convolve2d(x, y)
    if padding:
        return output[4:-4, 4:-4]
    else:
        return output

@jit
def fft_convolve2d(x, y):
    """ 
    2D convolution using the Fast Fourier Transform (FFT).

    :param x: first 2D array
    :param y: second 2D array

    :return: 2D array 

    """
    fr = fft2(x)
    fr2 = fft2(jnp.flipud(jnp.fliplr(y)))
    m, n = fr.shape
    cc = jnp.real(ifft2(fr * fr2))
    cc = jnp.roll(cc, int(-m / 2), axis=0)
    cc = jnp.roll(cc, int(-n / 2), axis=1)
    return cc


@jit
def scipy_convolve(data, kernel):
    """
    FFT-based Scipy convolution.

    :param data: first array 
    :param kernel: second array

    :return: output array

    """
    return jax.scipy.signal.convolve(data, kernel, mode='same', method='fft')

@jit
def gaussian_function(x, y, amp, sigma_x, sigma_y, center_x, center_y):
    """
    :param x: 1D array of x positions
    :param y: 1D array of y positions
    :param amp: the amplitude coefficient
    :type amp: float
    :param sigma_x: x-spread of the Gaussian blob
    :type sigma_x: float
    :param sigma_y: y-spread of the Gaussian blob
    :type sigma_y: float
    :param center_x: x position of the center
    :type center_x: float
    :param center_y: y position of the center
    :type center_y: float

    :return: 2D Gaussian

    """
    c = 1 / jnp.sqrt((2 * np.pi * sigma_x * sigma_y))
    delta_x = jnp.subtract(x, center_x)
    delta_y = jnp.subtract(y, center_y)
    exponent = -((delta_x / sigma_x) ** 2 + (delta_y / sigma_y) ** 2) / 2.
    g = c * jnp.exp(exponent)
    g /= g.sum()
    return amp * g

def make_grid(numPix, deltapix, subgrid_res=1):
    """
    Creates pixel grid as 1D arrays of x and y positions.
    The default coordinate frame is such that (0,0) is at the center of the coordinate grid.

    :param numPix: number of pixels per axis. Provide an integer for a square grid
    :param deltapix: pixel size
    :param subgrid_res: sub-pixel resolution 
    :return: x and y position information given as two 1D arrays

    """

    numPix = [numPix, numPix]

    # Super-resolution sampling
    # numPix_eff = int(numPix * subgrid_res)
    numPix_eff = [int(n * subgrid_res) for n in numPix]
    deltapix_eff = deltapix / float(subgrid_res)

    # Compute unshifted grids.
    # X values change quickly, Y values are repeated many times
    # NOTE jax.numpy.tile checks if `reps` is of type int, but numpy.int64
    #      is not in fact this type. Simply casting as int(numPix_eff[1])
    #      causes problems elsewhere with jax tracing, so we use another approach
    # x_grid = np.tile(np.arange(numPix_eff[0]), numPix_eff[1]) * deltapix_eff
    # y_grid = np.repeat(np.arange(numPix_eff[1]), numPix_eff[0]) * deltapix_eff
    x_space = jnp.arange(numPix_eff[0]) * deltapix_eff
    y_space = jnp.arange(numPix_eff[1]) * deltapix_eff
    x_grid, y_grid = jnp.meshgrid(x_space, y_space)
    x_grid, y_grid = x_grid.flatten(), y_grid.flatten()

    # Shift so (0, 0) is centered
    shift = jnp.array([deltapix_eff * (n - 1) / 2 for n in numPix_eff])

    return x_grid - shift[0], y_grid - shift[1]

@jit
def fwhm2sigma(fwhm):
    """
    Converts the FWHM (Full Width at Half Maximum) to the Gaussian sigma.

    :param fwhm: the full width at half maximum value
    :type fwhm: float
    :return: Gaussian standard deviation, `i.e.`, sqrt(var)

    """
    sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
    return sigma

@jit
def moffat_function(x, y, amp, fwhm, beta, center_x, center_y):
    """
    :param x: 1D array of x positions
    :param y: 1D array of y positions
    :param amp: the amplitude coefficient
    :type amp: float
    :param fwhm: the full width at half maximum value
    :type fwhm: float
    :param beta: the Moffat beta parameter
    :type beta: float
    :param center_x: x position of the center
    :type center_x: float
    :param center_y: y position of the center
    :type center_y: float

    :return: 2D Moffat
    
    """
    r0 = fwhm / (2. * jnp.sqrt(2. ** (1. / beta) - 1.))
    delta_x = jnp.subtract(x, center_x)
    delta_y = jnp.subtract(y, center_y)
    rr_gg = (delta_x / r0) ** 2 + (delta_y / r0) ** 2
    return amp * (1 + rr_gg) ** (-beta)


@jit
def moffat_elliptical_function(x, y, amp, fwhm_x, fwhm_y, phi, beta, center_x, center_y):
    """
    :param x: 1D array of x positions
    :param y: 1D array of y positions
    :param amp: the amplitude coefficient
    :type amp: float
    :param fwhm_x: the full width at half maximum value in the x direction
    :type fwhm_x: float
    :param fwhm_y: the full width at half maximum value in the y direction
    :type fwhm_y: float
    :param phi: orientation angle
    :type phi: float
    :param beta: the Moffat beta parameter
    :type beta: float
    :param center_x: x position of the center
    :type center_x: float
    :param center_y: y position of the center
    :type center_y: float

    :return: 2D Elliptical Moffat

    """
    r0_x = fwhm_x / (2. * jnp.sqrt(2. ** (1. / beta) - 1.))
    r0_y = fwhm_y / (2. * jnp.sqrt(2. ** (1. / beta) - 1.))
    delta_x = jnp.subtract(x, center_x)
    delta_y = jnp.subtract(y, center_y)
    A = (jnp.cos(phi) / r0_x)**2  + (jnp.sin(phi) / r0_y)**2
    B = (jnp.sin(phi) / r0_x)**2  + (jnp.cos(phi) / r0_y)**2
    C = 2*jnp.sin(phi)*jnp.cos(phi)*(1./ r0_x**2 - 1/r0_y**2)
    rr_gg = A*(delta_x ** 2) + B * (delta_y ** 2) + C * delta_x * delta_y
    return amp * (1 + rr_gg) ** (-beta)

def twoD_Gaussian(x,y, amplitude, xo, yo, sigma_x, sigma_y, theta):
    """
    Analytical 2D Gaussian function

    :param x: array containing x-axis variable
    :param y: array containing y-axis variable
    :param amplitude: amplitude of the Gaussian
    :param xo: center x coordinate
    :param yo: center y coordinate
    :param sigma_x: standard deviation of x variable
    :param sigma_y: standard deviation of y variable
    :param theta: ellongation angle
    """
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo)
                            + c*((y-yo)**2)))
    return g


def save_fits(array, path, header=None):
    """ 
    Saves ``.fits`` file of an array to a specified ``path``.
    
    """
    hdu = fits.PrimaryHDU(array, header=header)
    hdu.writeto(path + '.fits', overwrite=True)


def save_npy(array, path, header=None):
    """
    Saves ``.npy`` file of an array to a specified ``path``. Header is not used in this format.

    """
    np.save(path + '.npy', array)

def Downsample(image, factor=1, conserve_flux = False):
    """
    Resizes an image from dimensions (nx, ny) to (nx/factor, ny/factor).

    :param image: 2D array with shape (nx, ny)
    :param factor: downsampling factor, which must be greater than or equal to 1
    :type factor: int

    :return: 2D array

    """
    if factor == 1:
        return image
    if factor < 1:
        raise ValueError('scaling factor in re-sizing %s < 1' % factor)
    f = int(factor)
    nx, ny = np.shape(image)
    if int(nx / f) == nx / f and int(ny / f) == ny / f:
        small = image.reshape([int(nx / f), f, int(ny / f), f]).mean(3).mean(1)
        if conserve_flux:
            return small * f**2
        else:
            return small
    else:
        raise ValueError("scaling with factor %s is not possible with grid size %s, %s" % (f, nx, ny))


def Upsample(image, factor=1):
    """
    Resizes an image without interpolation.

    :param image: 2D array with shape (nx, ny)
    :param factor: upsampling factor, which must be greater than or equal to 1
    :type factor: int

    :return: 2D array

    """
    if factor == 1:
        return image
    if factor < 1:
        raise ValueError('scaling factor in re-sizing %s < 1' % factor)
    f = int(factor)
    n1, n2 = image.shape
    upimage = np.zeros((n1 * f, n2 * f))
    x, y = np.where(upimage == 0)
    x_, y_ = (x / f).astype(int), (y / f).astype(int)
    upimage[x, y] = image[x_, y_] / f ** 2
    return upimage


def convert_numpy_array_to_list(kwargs):
    new_dic = {}
    for k in kwargs.keys():
        dic_temp = {}
        for j in kwargs[k].keys():
            dic_temp[j] = deepcopy(np.asarray(kwargs[k][j]).tolist())
        new_dic[k] = deepcopy(dic_temp)

    return new_dic


def convert_list_to_numpy_array(kwargs):
    new_dic = {}
    for k in kwargs.keys():
        dic_temp = {}
        for j in kwargs[k].keys():
            if isinstance(kwargs[k][j], list):
                dic_temp[j] = deepcopy(np.asarray(kwargs[k][j]))
            else:
                dic_temp[j] = deepcopy(kwargs[k][j])
        new_dic[k] = deepcopy(dic_temp)

    return new_dic



import warnings

import jax.numpy as jnp
import numpy as np
from starred.utils.jax_utils import decompose, scale_norms


class Loss(object):
    """
    Class that manages the (auto-differentiable) loss function, defined as:
    L = - log(likelihood) - log(regularization)

    Note that gradient, hessian, etc. are computed in the ``InferenceBase`` class.

    """

    def __init__(self, data, psf_class, param_class, sigma_2, N, masks=None,
                 regularization_terms='l1_starlet', regularization_strength_scales=0, regularization_strength_hf=0,
                 regularization_strength_positivity=0, regularize_full_psf=True,
                 W=None):
        """
        :param data: array containing the observations
        :param psf_class: Point Spread Function (PSF) class from ``starred.psf.psf``
        :param param_class: parameters class from ``starred.psf.parameters``
        :param sigma_2: array containing the square of the noise maps
        :param N: number of observations stamps
        :type N: int
        :param masks: array containing the masks for the PSF (if given)
        :param regularization_terms: information about the regularization terms
        :type regularization_terms: str
        :param regularization_strength_scales: Lagrange parameter that weights intermediate scales in the transformed domain.
        :type regularization_strength_scales: float
        :param regularization_strength_hf: Lagrange parameter weighting the highest frequency scale
        :type regularization_strength_hf: float
        :param regularization_strength_positivity: Lagrange parameter weighting the positivity of the full PSF. 0 means no positivity constraint (recommended).
        :type regularization_strength_positivity: float
        :param regularize_full_psf: True if you want to regularize the Moffat and the background (recommended). False regularizes only the background
        :type regularize_full_psf: bool
        :param W: weight matrix. Shape (n_scale, n_pix*subsampling_factor, n_pix*subsampling_factor)

        """
        self._data = data
        self._psf = psf_class
        self._param = param_class
        self.N = N
        self._sigma_2 = sigma_2
        self.W = W
        self.regularize_full_psf = regularize_full_psf
        self._masks = masks

        self._init_likelihood()
        self._init_regularizations(regularization_terms, regularization_strength_scales, regularization_strength_hf,
                                   regularization_strength_positivity)

        self._penalty = 1e10

    def __call__(self, args):
        return self.loss(args)

    def loss(self, args):
        """Defined as the negative log(likelihood*regularization)"""
        kwargs = self._param.args2kwargs(args)
        neg_log = - self._log_likelihood(kwargs)
        if self._st_src_lambda != 0 or self._st_src_lambda_hf != 0:
            neg_log -= self._log_regul(kwargs)
        if self._pos_lambda !=0 :
            neg_log -= self._log_regul_positivity(kwargs)

        return jnp.nan_to_num(neg_log, nan=1e15, posinf=1e15, neginf=1e15)
     
    def update_dataset(self, newdata, newsigma2, newW, newparam_class):
        """Updates the dataset."""
        self._update_data(newdata, newsigma2)
        self._update_parameters(newparam_class)
        if newW is not None:
            self._update_weights(newW)
    
    def _update_parameters(self, param_class):
        """Updates the parameters."""
        self._param = param_class
        
    def _update_data(self, newdata, newsigma2):
        """Updates the data."""
        self._data = newdata
        self._sigma_2 = newsigma2

    def _update_weights(self, W):
        """Updates the weight matrix W."""
        self._st_src_norms = W[:-1]
        self.W = W

    @property
    def data(self):
        """Returns the observations array."""
        return self._data.astype(dtype=np.float32)

    @property
    def sigma_2(self):
        """Returns the noise map array."""
        return self._sigma_2.astype(dtype=np.float32)

    @property
    def masks(self):
        """Returns the masks array."""
        if self._masks is None:
            self._masks = np.ones((self.N, self.data.shape[1], self.data.shape[2]))
        return self._masks.astype(dtype=np.float32)

    def _init_likelihood(self):
        """Intialization of the data fidelity term of the loss function."""
        self._log_likelihood = self._log_likelihood_chi2

    def _init_regularizations(self, regularization_terms, regularization_strength_scales, regularization_strength_hf,
                              regularization_strength_positivity):
        """Intialization of the regularization terms of the loss function."""
        regul_func_list = []
        # add the log-regularization function to the list
        regul_func_list.append(getattr(self, '_log_regul_' + regularization_terms))

        if regularization_terms == 'l1_starlet':
            n_pix_src = min(*self.data[0, :, :].shape) * self._psf.upsampling_factor
            self.n_scales = int(np.log2(n_pix_src))  # maximum allowed number of scales
            if self.W is None:  # old fashion way
                if regularization_strength_scales != 0 and regularization_strength_hf != 0:
                    warnings.warn('lambda is not normalized. Provide the weight map !')
                wavelet_norms = scale_norms(self.n_scales)[:-1]  # ignore coarsest scale
                self._st_src_norms = jnp.expand_dims(wavelet_norms, (1, 2)) * jnp.ones(
                    (n_pix_src, n_pix_src))
            else:
                self._st_src_norms = self.W[:-1]  # ignore the coarsest scale
            self._st_src_lambda = float(regularization_strength_scales)
            self._st_src_lambda_hf = float(regularization_strength_hf)
        else:
            raise NotImplementedError(f'Regularisation term {regularization_terms} not implemented yet. Please use "l1_starlet". ')

        # positivity term
        self._pos_lambda = float(regularization_strength_positivity)
        # build the composite function (sum of regularization terms)
        self._log_regul = lambda kw: sum([func(kw) for func in regul_func_list])

    def _log_likelihood_chi2(self, kwargs):
        """Computes the data fidelity term of the loss function using chi2."""
        likelihood = - 0.5 * (1. / self.N) * jnp.sum(jnp.array([jnp.sum(
            jnp.multiply(self.masks[i, :, :], jnp.subtract(self.data[i, :, :], self._psf.model(i, **kwargs))) ** 2
            / self.sigma_2[i, :, :]) for i in range(self.N)]))
        return likelihood

    def _log_regul_l1_starlet(self, kwargs):
        """Computes the regularization terms as the sum of:
        the L1 norm of the Starlet transform of the highest frequency scale, and
        the L1 norm of the Starlet transform of all remaining scales (except the coarsest)."""
        if self.regularize_full_psf:
            h = self._psf.get_narrow_psf(**kwargs, norm=False)
        else:
            h = self._psf.get_background(kwargs['kwargs_background'])
        st = decompose(h, self.n_scales)[:-1]  # ignore the coarsest scale, which is a constant
        st_weighted_l1_hf = jnp.sum(self._st_src_norms[0] * jnp.abs(st[0]))  # first scale (i.e. high frequencies)
        st_weighted_l1 = jnp.sum(
            self._st_src_norms[1:] * jnp.abs(st[1:]))  # other scales, we ignore the coarsest scale in W as well
        l1_tot_regul = - (self._st_src_lambda_hf * st_weighted_l1_hf + self._st_src_lambda * st_weighted_l1)
        return l1_tot_regul / self._psf.upsampling_factor**2

    def _log_regul_positivity(self, kwargs):
        if self._pos_lambda > 0.:
            #penalise if the full PSF model contains <0 pixels (narrow PSF can have <0 pixels!)
            psf = self._psf.get_full_psf(**kwargs)
            sum_pos = -jnp.where(psf < 0., psf, 0.).sum()
            return - self._pos_lambda * sum_pos
        else:
            return 0.



# Utility methods for Sersic profiles
#
# Adapted from Herculens

import jax.numpy as jnp
from jax import lax

smoothing = 0.00001


def sersic_ellipse(x, y, R_sersic, n_sersic, e1, e2, center_x, center_y, amp, max_R_frac=100.0):
    """

    :param x:
    :param y:
    :param amp: surface brightness/amplitude value at the half light radius
    :param R_sersic: semi-major axis half light radius, in pixel
    :param n_sersic: Sersic index
    :param e1: eccentricity parameter
    :param e2: eccentricity parameter
    :param center_x: center in x-coordinate
    :param center_y: center in y-coordinate
    :param max_R_frac: truncation radius, in units of R_sersic (float)
    :return: Sersic profile value at (x, y)
    """

    R_sersic = jnp.maximum(0, R_sersic)
    phi_G, q = ellipticity2phi_q(e1, e2)
    center_x = lax.cond(center_x % 1. == 0.5, lambda x: x + 1e-4, lambda x: x, operand=center_x)
    center_y = lax.cond(center_y % 1. == 0.5, lambda x: x + 1e-4, lambda x: x,
                        operand=center_y)  # this is to avoid cancelling gradient if the sersic is exactly placed on the center of a pixel
    R = get_distance_from_center(x, y, phi_G, q, center_x, center_y)
    result = r_sersic(R, R_sersic, n_sersic, max_R_frac)
    return amp * result


def get_distance_from_center(x, y, phi_G, q, center_x, center_y):
    """
    Get the distance from the center of Sersic, accounting for orientation and axis ratio
    :param x:
    :param y:
    :param phi_G: orientation angle in rad
    :param q: axis ratio
    :param center_x: center x of sersic
    :param center_y: center y of sersic
    """
    x_shift = x - center_x
    y_shift = y - center_y
    cos_phi = jnp.cos(phi_G)
    sin_phi = jnp.sin(phi_G)
    xt1 = cos_phi * x_shift + sin_phi * y_shift
    xt2 = -sin_phi * x_shift + cos_phi * y_shift
    xt2difq2 = xt2 / (q * q)
    R = jnp.sqrt(xt1 * xt1 + xt2 * xt2difq2)
    return R


def R_stable(R):
    """
    Floor R_ at self._smoothing for numerical stability
    :param R: radius
    :return: smoothed and stabilized radius
    """
    return jnp.maximum(smoothing, R)


def r_sersic(R, R_sersic, n_sersic, max_R_frac=100.0):
    """

    :param R: radius (array or float)
    :param R_sersic: Sersic radius (half-light radius)
    :param n_sersic: Sersic index (float)
    :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
    :return: kernel of the Sersic surface brightness at R
    """
    # Must avoid item assignment on JAX arrays
    R_ = R_stable(R)
    R_sersic_ = R_stable(R_sersic)
    bn = b_n(n_sersic)
    R_frac = R_ / R_sersic_
    good_inds = (jnp.asarray(R_frac) <= max_R_frac).astype(int)
    result = good_inds * jnp.exp(-bn * (R_frac ** (1. / n_sersic) - 1.))
    return jnp.nan_to_num(result)


def b_n(n):
    """
    b(n) computation. This is the approximation of the exact solution to the relation, 2*incomplete_gamma_function(2n; b_n) = Gamma_function(2*n).
    :param n: the sersic index
    :return:
    """
    bn = 1.9992 * n - 0.3271
    return bn


def phi_q2_ellipticity(phi, q):
    """
    transforms orientation angle and axis ratio into complex ellipticity moduli e1, e2

    :param phi: angle of orientation (in radian)
    :param q: axis ratio minor axis / major axis
    :return: eccentricities e1 and e2 in complex ellipticity moduli
    """
    e1 = (1. - q) / (1. + q) * jnp.cos(2 * phi)
    e2 = (1. - q) / (1. + q) * jnp.sin(2 * phi)
    return e1, e2


def ellipticity2phi_q(e1, e2):
    """Transform complex ellipticity components to position angle and axis ratio.

    Parameters
    ----------
    e1, e2 : float or array_like
        Ellipticity components.

    Returns
    -------
    phi, q : same type as e1, e2
        Position angle (rad) and axis ratio (semi-minor / semi-major axis)

    """
    # replace value by low float instead to avoid NaNs
    # e1 = lax.cond(e1 == 0.0, lambda _: 1e-4, lambda _: e1, operand=None)  # does not work with TFP!
    # e2 = lax.cond(e2 == 0.0, lambda _: 1e-4, lambda _: e2, operand=None)  # does not work with TFP!
    e1 = jnp.where(e1 == 0., 1e-4, e1)
    e2 = jnp.where(e2 == 0., 1e-4, e2)
    phi = jnp.arctan2(e2, e1) / 2.
    c = jnp.sqrt(e1 ** 2 + e2 ** 2)
    c = jnp.minimum(c, 0.9999)
    q = (1. - c) / (1. + c)
    return phi, q



import os
import sys

import numpy as np
from astropy.wcs import WCS

"""
A set of functions to interact with ds9 region

"""


def export_astrometry_as_ds9region(model_deconv, outputfolder, kwargs, header=None, regsize=0.02):
    """"
    Export your final results as a DS9 readable region.

    :param model_deconv: starred.deconvolution.Deconv object.
    :param outputfolder: Path to the output folder.
    :param kwargs: dictionary containing the parameters of the model
    :param header: header of the fits file of the first epoch of the data, must contains WCS coordinate. If None, it will be exported in pixel image coordinate.
    :param regsize: Region size, in arcsecond or in pixel if the header is not provided.

    """
    final_cx = kwargs['kwargs_analytic']['c_x']
    final_cy = kwargs['kwargs_analytic']['c_y']

    if header is not None:
        wcs = WCS(header=header)
        radec_out = wcs.pixel_to_world(
            final_cx + model_deconv.image_size / 2. - 2 + 1. / model_deconv._upsampling_factor,
            final_cy + model_deconv.image_size / 2. - 2 + 1. / model_deconv._upsampling_factor)  # -1 to match DS9 indexing convention, other correction to match the subsampling resolution

        with open(os.path.join(outputfolder, 'point_sources_astrometry.reg'), 'w') as f:
            f.write(
                f'# Region file format: DS9 version 4.1 global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source={model_deconv.M} \nfk5 \n')
            for i in range(model_deconv.M):
                stri = radec_out.to_string('hmsdms', sep=':')[i].replace(' ', ',')
                f.write(f'circle({stri},{regsize}") \n')

    else:
        with open(os.path.join(outputfolder, 'point_sources_astrometry.reg'), 'w') as f:
            f.write(
                f'# Region file format: DS9 version 4.1 global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source={model_deconv.M} \nimage \n')
            for i in range(model_deconv.M):
                x = final_cx[i] + model_deconv.image_size / 2. + 0.5
                y = final_cy[i] + model_deconv.image_size / 2. + 0.5  # To match half pixel indexing of ds9
                f.write(f'circle({x}, {y}, {regsize}) \n')



import warnings

import jax.numpy as jnp
import numpy as np
from utils.jax_utils import decompose, scale_norms, get_l1_norm
from utils.generic_utils import gaussian_function
from copy import deepcopy
from jax.lax import dynamic_slice, dynamic_update_slice
import warnings


class Loss(object):
    """
    Class that manages the (auto-differentiable) loss function, defined as:
    L = - log(likelihood) - log(regularization) - log(positivity) - log(prior)

    Note that gradient, hessian, etc. are computed in the ``InferenceBase`` class.

    """

    _supported_ll = ('l2_norm')
    _supported_regul_source = ('l1_starlet', 'l1_starlet_pts_rw')

    def __init__(self, data, deconv_class, param_class, sigma_2,
                 regularization_terms='l1_starlet', regularization_strength_scales=1.0, regularization_strength_hf=1.0,
                 regularization_strength_pts_source=0., regularization_strength_positivity=0.,
                 regularization_strength_positivity_ps=0.,
                 regularize_full_model=False, W=None, prior=None):
        """
        :param data: array containing the observations
        :param deconv_class: deconvolution class from ``starred.deconvolution.deconvolution``
        :param param_class: parameters class from ``starred.deconvolution.parameters``
        :param sigma_2: array containing the square of the noise maps
        :param N: number of observations stamps
        :type N: int
        :param regularization_terms: Choose the type of regularisation 'l1_starlet' or 'l1_starlet_pts_rw'
        :type regularization_terms: str
        :param regularization_strength_scales: Lagrange parameter that weights intermediate scales in the transformed domain 
        :type regularization_strength_scales: float
        :param regularization_strength_hf: Lagrange parameter weighting the highest frequency scale
        :type regularization_strength_hf: float
        :param regularization_strength_positivity: Lagrange parameter weighting the positivity of the background. 0 means no positivity constrain.
        :type regularization_strength_positivity: float
        :param regularization_strength_positivity_ps: Lagrange parameter weighting the positivity of the point sources. 0 means no positivity constrain.
        :type regularization_strength_positivity: float
        :param regularization_strength_pts_source: Lagrange parameter regularising the point source channel.
        :type regularization_strength_pts_source: float
        :param W: weight matrix. Shape (n_scale, n_pix*subsampling_factor, n_pix*subsampling_factor)
        :type W: jax.numpy.array
        :param regularize_full_model: option to regularise just the background (False) or background + point source channel (True)
        :type regularize_full_model: bool
        :param prior: Prior class containing Gaussian prior on the free parameters
        :type prior: Prior class

        """
        self._data = data
        self._sigma_2 = sigma_2
        self.W = W
        self._deconv = deconv_class
        self._datar = self._data.reshape(self._deconv.epochs,
                                         self._deconv.image_size,
                                         self._deconv.image_size)
        self._sigma_2r = self._sigma_2.reshape(self._deconv.epochs,
                                               self._deconv.image_size,
                                               self._deconv.image_size)

        self._param = param_class
        self.epochs = self._deconv.epochs
        self.regularize_full_model = regularize_full_model
        self.prior = prior

        self._init_likelihood()
        self._init_prior()
        self._init_regularizations(regularization_terms, regularization_strength_scales, regularization_strength_hf,
                                   regularization_strength_positivity, regularization_strength_positivity_ps,
                                   regularization_strength_pts_source)

    # @partial(jit, static_argnums=(0,))
    def __call__(self, args):
        return self.loss(args)

    def loss(self, args):
        """Defined as the negative log(likelihood*regularization)."""
        kwargs = self._param.args2kwargs(args)
        neg_log = - self._log_likelihood(kwargs)
        if self._st_src_lambda != 0 or self._st_src_lambda_hf !=0 :
            neg_log -= self._log_regul(kwargs)
        if self._pos_lambda != 0:
            neg_log -= self._log_regul_positivity(kwargs)
        if self._pos_lambda_ps != 0.:
            neg_log -= self._log_regul_positivity_ps(kwargs)
        if self.prior is not None:
            neg_log -= self._log_prior(kwargs)
        if self._lambda_pts_source != 0.:
            neg_log -= self._log_regul_pts_source(kwargs)

        return jnp.nan_to_num(neg_log, nan=1e15, posinf=1e15, neginf=1e15)

    @property
    def datar(self):
        """Returns the observations array."""
        return self._datar.astype(dtype=np.float32)

    @property
    def sigma_2r(self):
        """Returns the noise map array."""
        return self._sigma_2r.astype(dtype=np.float32)

    def _init_likelihood(self):
        """Intialization of the data fidelity term of the loss function."""
        self._log_likelihood = self._log_likelihood_chi2

    def _init_prior(self):
        """
        Initialization of the prior likelihood
        """
        if self.prior is None:
            self._log_prior =  lambda x: 0.
        else :
            self._log_prior = self._log_prior_gaussian

    def _init_regularizations(self, regularization_terms, regularization_strength_scales, regularization_strength_hf,
                              regularization_strength_positivity, regularization_strength_positivity_ps,
                              regularization_strength_pts_source):
        """Intialization of the regularization terms of the loss function."""
        regul_func_list = []
        # add the log-regularization function to the list
        regul_func_list.append(getattr(self, '_log_regul_' + regularization_terms))

        if regularization_terms == 'l1_starlet' or regularization_terms == 'l1_starlet_pts_rw':
            n_pix_src = min(*self.datar[0, :, :].shape) * self._deconv._upsampling_factor
            self.n_scales = int(np.log2(n_pix_src))  # maximum allowed number of scales
            if self.W is None:  # old fashion way
                if regularization_strength_scales != 0 and regularization_strength_hf != 0:
                    warnings.warn('lambda is not normalized. Provide the weight map !')
                wavelet_norms = scale_norms(self.n_scales)[:-1]  # ignore coarsest scale
                self._st_src_norms = jnp.expand_dims(wavelet_norms, (1, 2)) * jnp.ones((n_pix_src, n_pix_src))
            else:
                self._st_src_norms = self.W[:-1]  # ignore the coarsest scale
            self._st_src_lambda = float(regularization_strength_scales)
            self._st_src_lambda_hf = float(regularization_strength_hf)

        # compute l1 norm of a pts source of amp = 1
        self.l1_pts = self._init_l1_pts()

        # positivity term
        self._pos_lambda = float(regularization_strength_positivity)
        self._pos_lambda_ps = float(regularization_strength_positivity_ps)

        # regularisation strenght of the pts source channel
        self._lambda_pts_source = float(regularization_strength_pts_source)

        # build the composite function (sum of regularization terms)
        self._log_regul = lambda kw: sum([func(kw) for func in regul_func_list])

    def _init_l1_pts(self):
        one_pts = self._deconv.shifted_gaussians([0.], [0.], [1.], source_ID=[0])
        pts_im_starlet = decompose(one_pts, self.n_scales)[:-1]  # ignore coarsest scale
        return jnp.sum(self._st_src_norms * jnp.abs(pts_im_starlet))

    def _log_likelihood_chi2(self, kwargs):
        """Computes the data fidelity term of the loss function using the L2 norm."""
        model = self._deconv.model(kwargs)
        return - 0.5 * (1. / self.epochs) * jnp.sum((model - self.datar) ** 2 / self.sigma_2r)

        # return - 0.5 * jnp.sum(jnp.array([jnp.sum(jnp.subtract(self.data[i,:,:], self._deconv.model(i, **kwargs)[0])**2 / self.sigma_2[i,:,:]) for i in range(self.epochs)]))

    # @partial(jit, static_argnums=(0,))
    def _log_regul_l1_starlet(self, kwargs):
        """
        Computes the regularization terms as the sum of:
        
        - the L1 norm of the Starlet transform of the highest frequency scale, and
        - the L1 norm of the Starlet transform of all remaining scales (except the coarsest).
        """
        if self.regularize_full_model:
            toreg, _ = self._deconv.getDeconvolved(kwargs, epoch=None)
        else :
            _, toreg = self._deconv.getDeconvolved(kwargs, epoch=None)
        st = decompose(toreg, self.n_scales)[:-1]  # ignore coarsest scale
        st_weighted_l1_hf = jnp.sum(self._st_src_norms[0] * jnp.abs(st[0]))  # first scale (i.e. high frequencies)
        st_weighted_l1 = jnp.sum(self._st_src_norms[1:] * jnp.abs(st[1:]))  # other scales
        tot_l1_reg = - (self._st_src_lambda_hf * st_weighted_l1_hf + self._st_src_lambda * st_weighted_l1)

        return tot_l1_reg / self._deconv._upsampling_factor**2

    def _log_regul_positivity(self, kwargs):
        """
        Computes the posivity constraint term. A penalty is applied if the epoch with the smallest background mean has negative pixels.

        :param kwargs:
        """
        h = self._deconv._get_background(kwargs)
        sum_pos = -jnp.where(h < 0., h, 0.).sum()
        return - self._pos_lambda * sum_pos

    def _log_regul_positivity_ps(self, kwargs):
        """
        Computes the posivity constraint term for the point sources. A penalty is applied if one of the point sources have negative amplitude.

        :param kwargs:
        """
        fluxes = jnp.array(kwargs['kwargs_analytic']['a'])
        sum_pos = -jnp.where(fluxes < 0., fluxes, 0.).sum()
        return - self._pos_lambda * sum_pos

    def _log_regul_pts_source(self, kwargs):
        """
        Penalty term to the pts source, to compensate for the fact that the pts source channel is not regularized

        :param kwargs: dictionary containing all keyword arguments
        """
        pts_source_channel = self._deconv.getPts_source(kwargs, epoch=None)
        st = decompose(pts_source_channel, self.n_scales)[:-1]  # ignore coarsest scale
        st_weighted_l1 = jnp.sum(self._st_src_norms * jnp.abs(st))  # other scales
        tot_l1_reg = - self._lambda_pts_source * st_weighted_l1

        return tot_l1_reg / self._deconv._upsampling_factor**2

    def _log_prior_gaussian(self, kwargs):
        """
        A flexible term in the likelihood to impose a Gaussian prior on any parameters of the model.

        :param kwargs: dictionary containing all keyword arguments
        """
        return self.prior.logL(kwargs)

    def update_weights(self, W):
        """Updates the weight matrix W."""

        self._st_src_norms = W[:-1]
        self.W = W

    def update_lambda_pts_source(self, kwargs):
        """Update lambda_pts regularisation according to the new model, according to Appendix C of Millon et al. (2024).

        :param kwargs: dictionary containing all keyword arguments
        """

        deconvolved, background = self._deconv.getDeconvolved(kwargs)
        l1p_sum = 0
        for i in range(self._deconv.M):
            pts_im = self._deconv.getPts_source(kwargs, epoch=None, source_ID=[i])
            l1p_sum += - get_l1_norm(pts_im, self.W[:-1], self.n_scales) / self._deconv._upsampling_factor ** 2

        l1_bkg = - get_l1_norm(background, self.W[:-1], self.n_scales) / self._deconv._upsampling_factor ** 2
        l1_f = - get_l1_norm(deconvolved, self.W[:-1], self.n_scales) / self._deconv._upsampling_factor ** 2

        new_lambda_pts_source = float(self._st_src_lambda * (l1_f - l1_bkg) / l1p_sum)
        self._lambda_pts_source = new_lambda_pts_source

        return new_lambda_pts_source


class Prior(object):
    def __init__(self, prior_analytic=None, prior_background=None, prior_sersic=None):
        """

        :param prior_analytic: list of [param_name, mean, 1-sigma priors]
        :param prior_background: list of [param_name, mean, 1-sigma priors]
        :param prior_sersic: list of [param_name, mean, 1-sigma priors]
        """
        self._prior_analytic = prior_analytic
        self._prior_background = prior_background
        self._prior_sersic = prior_sersic

    def logL(self, kwargs):
        logL = 0
        logL += self.prior_kwargs(kwargs['kwargs_analytic'], self._prior_analytic)
        logL += self.prior_kwargs(kwargs['kwargs_background'], self._prior_background)
        logL += self.prior_kwargs(kwargs['kwargs_sersic'], self._prior_sersic)

        return logL

    @staticmethod
    def prior_kwargs(kwargs, prior):
        """

        :param kwargs: keyword argument
        :param prior: prior
        :return: logL
        """
        if prior is None:
            return 0
        logL = 0
        for i in range(len(prior)):
            param_name, value, sigma = prior[i]
            model_value = kwargs[param_name]
            dist = (model_value - value) ** 2 / sigma ** 2 / 2
            logL -= np.sum(dist)
        return logL




import time
import warnings

import jax
import numpy as np
import optax
from starred.optim.inference_base import FisherCovariance
from tqdm import tqdm
import jaxopt

from starred.optim.inference_base import InferenceBase


__all__ = ['Optimizer']

SUPPORTED_METHOD_OPTAX = ['adam', 'adabelief', 'radam']
SUPPORTED_METHOD_JAXOPT = ['BFGS', 'GradientDescent', 'LBFGS', 'l-bfgs-b', 'LevenbergMarquardt']
SUPPORTED_METHOD_SCIPYMINIMIZE = ['Nelder-Mead', 'Powell', 'CG', 'Newton-CG', 'TNC', 'COBYLA', 'SLSQP', 'trust-constr',
                                  'dogleg', 'trust-ncg', 'trust-exact', 'trust-krylov']
SUPPORTED_BOUNDED_METHOD = ['l-bfgs-b']

# GRADIENT_METHOD = ['Nelder-Mead', 'Powell', 'CG' 'Newton-CG', 'trust-krylov', 'trust-exact', 'trust-constr']
# HVP_METHOD = ['Nelder-Mead', 'Powell', 'CG','Newton-CG', 'trust-constr', 'trust-krylov', 'trust-constr']


class Optimizer(InferenceBase):
    """
        Class that handles optimization tasks, `i.e.`, finding best-fit point estimates of parameters
        It currently handles a subset of scipy.optimize.minimize routines,
        using first and second order derivatives when required.
    """

    def __init__(self, loss_class, param_class, method='BFGS'):
        super().__init__(loss_class, param_class)
        if method not in SUPPORTED_METHOD_OPTAX + SUPPORTED_METHOD_JAXOPT + SUPPORTED_METHOD_SCIPYMINIMIZE :
            raise NotImplementedError(f'Minimisation method {method} is not supported yet.')
        else:
            self.method = method

        self._metrics = MinimizeMetrics(self.loss, self.method)
    @property
    def loss_history(self):
        """Returns the loss history."""
        if len(self._metrics.loss_history) == 0:
            raise ValueError("You must run the optimizer at least once to access the history")
        return self._metrics.loss_history

    @property
    def param_history(self):
        """Returns the parameter history."""
        if len(self._metrics.param_history) == 0:
            raise ValueError("You must run the optimizer at least once to access the history")
        return self._metrics.param_history

    def minimize(self, restart_from_init=True, **kwargs_optimiser):
        """Minimizes the loss function and returns the best fit."""
        init_params = self._param.current_values(as_kwargs=False, restart=restart_from_init, copy=True)
        start = time.time()

        if self.method in SUPPORTED_METHOD_OPTAX:
            best_fit, extra_fields = self._run_optax(init_params, **kwargs_optimiser)
        elif self.method in SUPPORTED_METHOD_JAXOPT + SUPPORTED_METHOD_SCIPYMINIMIZE:
            best_fit, extra_fields = self._run_jaxopt(init_params, **kwargs_optimiser)
        else:
            raise NotImplementedError(f'Minimization method {self.method} is not supported yet.')

        runtime = time.time() - start
        self._param.set_best_fit(best_fit)
        logL_best_fit = self.loss(best_fit)

        return best_fit, logL_best_fit, extra_fields, runtime

    def _run_jaxopt(self, params, **kwargs_optimiser):
        extra_kwargs = {}
        if self.method in SUPPORTED_BOUNDED_METHOD:
            extra_kwargs['bounds'] = self.parameters.get_bounds()
        else:
            warnings.warn('You are using an unconstrained optimiser. Bounds are ignored.')

        if self.method == 'BFGS':
            solver = jaxopt.BFGS(fun=self.loss, **kwargs_optimiser)
        elif self.method == 'GradientDescent':
            solver = jaxopt.GradientDescent(fun=self.loss, **kwargs_optimiser)
        elif self.method == 'LBFGS':
            solver = jaxopt.LBFGS(fun=self.loss, **kwargs_optimiser)
        elif self.method in SUPPORTED_METHOD_SCIPYMINIMIZE :
            solver = jaxopt.ScipyMinimize(fun=self.loss, method=self.method, callback=self._metrics,
                                          **kwargs_optimiser)
        elif self.method == 'l-bfgs-b':
            solver = jaxopt.ScipyBoundedMinimize(fun=self.loss, method="l-bfgs-b", callback=self._metrics,
                                                 **kwargs_optimiser)
        elif self.method == 'LevenbergMarquardt':
            solver = jaxopt.LevenbergMarquardt(residual_fun=lambda x: jax.numpy.array([self.loss(x)]), 
                                               **kwargs_optimiser)

        res = solver.run(params, **extra_kwargs)
        extra_fields = {'result_class': res, 'stat':res.state, 'loss_history':self._metrics.loss_history}

        return res.params, extra_fields

    def _run_optax(self, params, max_iterations=100, min_iterations=None,
                   init_learning_rate=1e-2, schedule_learning_rate=True,
                   restart_from_init=False, stop_at_loss_increase=False,
                   progress_bar=True, return_param_history=False, decay_rate=0.99):

        if min_iterations is None:
            min_iterations = max_iterations
        if schedule_learning_rate is True:
            # Exponential decay of the learning rate
            scheduler = optax.exponential_decay(
                init_value=init_learning_rate,
                decay_rate=decay_rate,
                transition_steps=max_iterations)

            if self.method == 'adabelief':
                scale_algo = optax.scale_by_belief()
            elif self.method == 'radam':
                scale_algo = optax.scale_by_radam()
            elif self.method == 'adam':
                scale_algo = optax.scale_by_adam()
            else:
                raise NotImplementedError(f"Optax algorithm '{self.method}' is not supported")

            # Combining gradient transforms using `optax.chain`
            optim = optax.chain(
                # optax.clip_by_global_norm(1.0),  # clip by the gradient by the global norm
                scale_algo,  # use the updates from the chosen optimizer
                optax.scale_by_schedule(scheduler),  # Use the learning rate from the scheduler
                optax.scale(-1.)  # because gradient *descent*
            )
        else:
            if self.method == 'adabelief':
                optim = optax.adabelief(init_learning_rate)
            elif self.method== 'radam':
                optim = optax.radam(init_learning_rate)
            elif self.method == 'adam':
                optim = optax.adam(init_learning_rate)
            else:
                raise NotImplementedError(f"Optax algorithm '{self.method}' is not supported")

        # Initialise optimizer state
        opt_state = optim.init(params)
        prev_params, prev_loss_val = params, 1e10

        @jax.jit
        def gradient_step(params, opt_state):
            # loss_val, grads = jax.value_and_grad(self._loss)(params)
            loss_val, grads = self.value_and_gradient(params)
            updates, opt_state = optim.update(grads, opt_state, params)
            params = optax.apply_updates(params, updates)
            return params, opt_state, loss_val

        # Gradient descent loop
        for i in self._for_loop(range(int(max_iterations)), progress_bar,
                                total=int(max_iterations),
                                desc=f"optax.{self.method}"):
            params, opt_state, loss_val = gradient_step(params, opt_state)
            if stop_at_loss_increase and i > min_iterations and loss_val > prev_loss_val: #pragma: no cover
                params, loss_val = prev_params, prev_loss_val
                break
            else:
                self._metrics.loss_history.append(loss_val)
                prev_params, prev_loss_val = params, loss_val
            if return_param_history is True:
                self._metrics.param_history.append(params)

        best_fit = params
        extra_fields = {'loss_history': np.array(
            self._metrics.loss_history)}

        if return_param_history is True:
            extra_fields['param_history'] = self._metrics.param_history

        return best_fit, extra_fields

    @staticmethod
    def _for_loop(iterable, progress_bar_bool, **tqdm_kwargs):
        if progress_bar_bool is True:
            return tqdm(iterable, **tqdm_kwargs)
        else:
            return iterable

    def compute_fisher_matrix(self):
        if len(self._param.best_fit_values(as_kwargs=False)) > 500:
            warnings.warn('Computing the Fisher matrix for more than 500 dimensions will blow up your memory. '
                          'You should reduce the dimensionnality of your model')

        fish = FisherCovariance(self._param, self)
        fish.compute_fisher_information()

        return fish


class MinimizeMetrics(object):
    """Simple callable class used as callback in ``scipy.optimize.minimize`` method."""

    def __init__(self, func, method):
        self.loss_history = []
        self.param_history = []
        self._func = func
        if method == 'trust-constr':
            self._call = None #bug fix for now, jaxopt does not support 2 argument callback for now : https://github.com/google/jaxopt/issues/428
            # self._call = self._call_2args
        else:
            self._call = self._call_1arg

    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)

    def _call_1arg(self, x):
        self.loss_history.append(self._func(x))
        self.param_history.append(x)

    def _call_2args(self, x, state):
        # Input state parameter is necessary for 'trust-constr' method
        # You can use it to stop execution early by returning True
        self.loss_history.append(self._func(x))
        self.param_history.append(x)



import os
import dill as pkl
import warnings
from packaging import version
import h5py
import json

import jax.numpy as jnp
import numpy as np
import jax
from jax import lax

from starred.utils.generic_utils import pad_and_convolve, pad_and_convolve_fft, fwhm2sigma, gaussian_function, \
    make_grid, \
    moffat_elliptical_function, moffat_function, Downsample, scipy_convolve, save_npy, save_fits, \
    convert_numpy_array_to_list, convert_list_to_numpy_array


class PSF(object):
    """
    Narrow Point Spread Function class. Coordinates and FWHM are given in the original pixel grid.
    
    """

    def __init__(
            self,
            image_size=64,
            number_of_sources=2,
            upsampling_factor=2,
            gaussian_fwhm=2,
            convolution_method='scipy',
            gaussian_kernel_size=None,
            include_moffat=True,
            elliptical_moffat=False,
    ):
        """
        :param image_size: input images size in pixels
        :type image_size: int
        :param number_of_sources: amount of input images (each containing a point source)
        :type number_of_sources: int
        :param upsampling_factor: the rate at which the sampling frequency increases in the PSF with respect to the input images
        :type upsampling_factor: int
        :param gaussian_fwhm: the Gaussian's FWHM in pixels. Default is 2.
        :type gaussian_fwhm: int
        :param convolution_method: method to use to calculate the convolution, choose between 'fft', 'scipy', and 'lax. Recommended if jax>=0.4.9 : 'scipy'
        :type convolution_method: str
        :param gaussian_kernel_size: dimensions of the Gaussian kernel, not used if method = 'fft'. None will select the recommended size for each method.
        :type gaussian_kernel_size: int
        :param include_moffat: True for the PSF to be expressed as the sum of a Moffat and a grid of pixels. False to not include the Moffat. Default: True
        :type include_moffat: bool
        :param elliptical_moffat: Allow elliptical Moffat.
        :type elliptical_moffat: bool

        """
        self.elliptical_moffat = elliptical_moffat
        if self.elliptical_moffat:
            self.analytic = self._analytic_ellitpical
        else:
            self.analytic = self._analytic_circular

        self.image_size = image_size
        self.upsampling_factor = upsampling_factor
        self.image_size_up = self.image_size * self.upsampling_factor
        self.M = number_of_sources
        self.include_moffat = include_moffat

        if convolution_method == 'fft':
            self._convolve = pad_and_convolve_fft
            self.gaussian_size = self.image_size_up
            if (self.image_size * self.upsampling_factor) % 2 == 1:
                self.shift_gaussian = 0.
            else :
                self.shift_gaussian = 0.5 #convention for even kernel size and fft convolution
            self.shift_direction = 1.
        elif convolution_method == 'lax':
            self._convolve = pad_and_convolve
            if gaussian_kernel_size is None:
                self.gaussian_size = 12
            else:
                self.gaussian_size = gaussian_kernel_size
            self.shift_gaussian = 0.5
            self.shift_direction = 1.
        elif convolution_method == 'scipy':
            self._convolve = scipy_convolve
            if gaussian_kernel_size is None:
                self.gaussian_size = 16
            else:
                self.gaussian_size = gaussian_kernel_size
            if version.parse(jax.__version__) < version.parse('0.4.9'):
                warnings.warn("WARNING : jax.scipy has no FFT implementation for the moment. It might be faster to use 'fft' on CPU.")

            self.shift_gaussian = 0.5
            self.shift_direction = -1 #scipy has inverted shift convention
        else:
            raise NotImplementedError('Unknown convolution method : choose fft, scipy or lax')

        self.convolution_method = convolution_method
        self.fwhm = gaussian_fwhm
        self.sigma = fwhm2sigma(self.fwhm)

    def shifted_gaussians(self, i, x0, y0):
        """
        Generates a 2D array with point sources. Normalized to 1.

        :param i: current point source index 
        :type i: int
        :param x0: 1D array containing the x positions of the point sources, shift are given in unit of 'big' pixels
        :param y0: 1D array containing the y positions of the point sources, shift are given in unit of 'big' pixels
        :return: 2D array

        """
        x, y = make_grid(numPix=self.gaussian_size, deltapix=1.)
        return jnp.array(gaussian_function(
            x=x, y=y,
            amp=1, sigma_x=self.sigma,
            sigma_y=self.sigma,
            center_x=-(self.upsampling_factor * x0[i])*self.shift_direction - self.shift_gaussian,
            center_y=-(self.upsampling_factor * y0[i])*self.shift_direction - self.shift_gaussian,  # Convention : gaussian is placed at a center of a pixel
        ).reshape(self.gaussian_size, self.gaussian_size))


    def _analytic_ellitpical(self, fwhmx, fwhmy, phi, beta):
        """
        Generates the narrow PSF's analytic term.

        :param fwhm_x: the full width at half maximum value in the x direction
        :type fwhm_x: float
        :param fwhm_y: the full width at half maximum value in the y direction
        :type fwhm_y: float
        :param phi: orientation angle
        :type phi: float
        :param beta: the Moffat beta parameter
        :type beta: float

        :return: 2D Moffat

        """
        x, y = make_grid(numPix=self.image_size_up, deltapix=1.)
        return jnp.array(moffat_elliptical_function(
            x=x, y=y,
            amp=1, fwhm_x=fwhmx * self.upsampling_factor,
            fwhm_y = fwhmy * self.upsampling_factor, phi=phi, beta=beta,
            center_x=0,
            center_y=-0,
        ).reshape(self.image_size_up, self.image_size_up))

    def _analytic_circular(self, fwhm, beta):
        """
        Generates the narrow PSF's analytic term.

        :param fwhm: the full width at half maximum value in the x direction
        :type fwhm: float
        :param beta: the Moffat beta parameter
        :type beta: float

        :return: 2D Moffat

        """
        x, y = make_grid(numPix=self.image_size_up, deltapix=1.)
        return jnp.array(moffat_function(
            x=x, y=y,
            amp=1, fwhm=fwhm * self.upsampling_factor,
            beta=beta,
            center_x=0,
            center_y=-0,
        ).reshape(self.image_size_up, self.image_size_up))

    # @partial(jit, static_argnums=(0, 1, 5))
    def model(self, init=0, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None, high_res=False):
        """
        Creates the 2D narrow Point Spread Function (PSF) image.
        
        :param init: stamp index 
        :type init: int
        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels
        :param high_res: returns the upsampled version of the PSF
        :type high_res: bool
        :return: array containing the model
        
        """

        s = self.get_narrow_psf(kwargs_moffat, kwargs_gaussian, kwargs_background, norm=False)
        r = self.get_gaussian(init, kwargs_gaussian, kwargs_moffat)
        if high_res:
            model = self._convolve(s, r).reshape(self.image_size_up, self.image_size_up)
        else:
            model = Downsample(self._convolve(s, r).reshape(self.image_size_up, self.image_size_up),
                               factor=self.upsampling_factor)

        return model

    def get_moffat(self, kwargs_moffat, norm=True):
        """
        Returns the analytical part of the PSF.

        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param norm: normalizes the Moffat
        :type norm: bool

        """
        if self.elliptical_moffat:
            moff = kwargs_moffat['C'] * self.analytic(kwargs_moffat['fwhm_x'],kwargs_moffat['fwhm_y'],kwargs_moffat['phi'], kwargs_moffat['beta'])
        else:
            moff = kwargs_moffat['C'] * self.analytic(kwargs_moffat['fwhm'], kwargs_moffat['beta'])
        if norm:
            moff = moff / moff.sum()

        return moff

    def get_narrow_psf(self, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None, norm=True):
        """
        Returns the `narrow` PSF s.

        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels
        :return: array containing the `narrow` PSF
        """

        background = self.get_background(kwargs_background)
        if self.include_moffat:
            moff = self.get_moffat(kwargs_moffat, norm=False)
            s = moff + background
        else:
            s = background
            moff = jnp.zeros_like(s)
            s += lax.cond(s.sum() == 0, lambda _: 1e-6, lambda _: 0., operand=None)  # to avoid division per 0

        if norm:
            s = s / s.sum()

        return s

    def get_full_psf(self, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None, norm=True, high_res=True):
        """
        Returns the PSF of the original image, `i.e.`, the convolution between s, the `narrow` PSF, and r, the Gaussian kernel.

        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels
        :return: array containing the full PSF
        
        """
        s = self.get_narrow_psf(kwargs_moffat=kwargs_moffat, kwargs_gaussian=kwargs_gaussian, kwargs_background=kwargs_background, norm=norm)
        r = self.shifted_gaussians(0, [0.], [0.])
        if high_res:
            psf = self._convolve(s, r)
        else :
            psf = Downsample(self._convolve(s, r), factor=self.upsampling_factor)
        if norm:
            psf /= psf.sum()

        return psf

    def get_background(self, kwargs_background):
        """
        Returns the numerical part of the PSF.

        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels
        :return: array containing the background correction
        """

        return kwargs_background['background'].reshape(self.image_size_up, self.image_size_up)

    def get_gaussian(self, init, kwargs_gaussian, kwargs_moffat):
        """
        Returns a Gaussian function, adjusted to the star of index ``init``.

        :param init: stamp index
        :type init: int
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :return: array containing the Gaussian kernel
        """
        x0, y0 = kwargs_gaussian['x0'], kwargs_gaussian['y0']
        r = self.shifted_gaussians(init, x0, y0)

        # ga = kwargs_gaussian['a'][init] / jnp.mean(kwargs_gaussian['a']) #normalization by the mean to remove degeneracy with the C parameter of the moffat
        ga = kwargs_gaussian['a'][init] / kwargs_moffat[
            'C']  # normalization by the mean to remove degeneracy with the C parameter of the moffat
        r = ga * r

        return r

    def get_amplitudes(self, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None):
        """
        Returns the photometry of the stars.

        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels

        :return: list containing the relative photometry
       """

        kernel = self.get_narrow_psf(kwargs_moffat, kwargs_gaussian, kwargs_background, norm=False)
        kernel_norm = kernel.sum()
        amp = (kwargs_gaussian['a'] / kwargs_moffat['C']) * kernel_norm
        return amp

    def get_photometry(self, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None, high_res=False):
        """
        Returns the PSF photometry of all the stars.

        :param init: stamp index
        :type init: int
        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels
        :return: array containing the photometry
        """

        return jnp.array(
            [self.model(i, kwargs_moffat, kwargs_gaussian, kwargs_background, high_res).sum() for i in range(self.M)])

    def get_astrometry(self, kwargs_moffat=None, kwargs_gaussian=None, kwargs_background=None):
        """
        Returns the astrometry. In units of 'big' pixel.

        :param kwargs_moffat: dictionary containing keyword arguments corresponding to the analytic term of the PSF
        :param kwargs_gaussian: dictionary containing keyword arguments corresponding to the Gaussian deconvolution kernel
        :param kwargs_background: dictionary containing keyword arguments corresponding to the grid of pixels

        :return: list of tuples with format [(x1,y1), (x2,y2), ...]
        """
        coord = []
        for x, y in zip(kwargs_gaussian['x0'], kwargs_gaussian['y0']):
            coord.append([x, y])

        return np.asarray(coord)

    def export(self, output_folder, kwargs_final, data, sigma_2, format='fits'):
        """
        Saves all the output files in fits or npy format.

        :param output_folder: path to the output folder
        :type output_folder: str
        :param kwargs_final: dictionary containing all keyword arguments
        :param data: array containing the images
        :param sigma_2: array containing the noise maps
        :param format: output format. Choose between ``npy`` or ``fits``
        :type format: str
        """
        if format == 'fits':
            save_fct = save_fits
        elif format == 'npy':
            save_fct = save_npy
        else:
            raise NotImplementedError(f'Format {format} unknown.')

        narrow = self.get_narrow_psf(**kwargs_final, norm=True)
        save_fct(narrow, os.path.join(output_folder, 'narrow_PSF'))

        full = self.get_full_psf(**kwargs_final, norm=True, high_res=True)
        save_fct(full, os.path.join(output_folder, 'full_PSF'))

        background = self.get_background(kwargs_final['kwargs_background'])
        save_fct(background, os.path.join(output_folder, 'background_PSF'))

        analytic = self.get_moffat(kwargs_final['kwargs_moffat'], norm=True)
        save_fct(analytic, os.path.join(output_folder, 'analytic_PSF'))

        for i in range(self.M):
            estimated_full_psf = self.model(i, **kwargs_final)
            dif = data[i, :, :] - estimated_full_psf
            rr = jnp.abs(dif) / jnp.sqrt(sigma_2[i, :, :])

            save_fct(estimated_full_psf, os.path.join(output_folder, f'full_psf_{i}'))
            save_fct(dif, os.path.join(output_folder, f'residuals_{i}'))
            save_fct(rr, os.path.join(output_folder, f'scaled_residuals_{i}'))

    def dump(self, path, kwargs, norm, data=None, sigma_2=None, masks=None, save_output_level=4, format='hdf5'):
        """
        Stores information in a given file in pickle or hdf5 format (recommended).

        :param path: Filename of the output.
        :param kwargs: Dictionary containing the fitted value of the model
        :param norm: Normalisation factor of your data. This is an important to save if you want to get the correct photometry.
        :param data: (Nstar x image_size x image_size) array containing the data
        :param sigma_2: (Nstar x image_size x image_size) array containing the noise maps
        :param masks: (Nstar x image_size x image_size) array containing the noise maps
        :param save_output_level: Int. Level of output product to save: 1-just the parameters of the model, 2- add the input data, 3- add the output products (background, narrow PSF, full PSF) 4- add the output products for every image.

        """
        if format == 'pkl':
            with open(path, 'wb') as f:
                pkl.dump([self, kwargs, norm], f, protocol=pkl.HIGHEST_PROTOCOL)
        elif format == 'hdf5':
            kwargs_model = {
                'image_size': int(self.image_size),
                'number_of_sources': int(self.M),
                'upsampling_factor': int(self.upsampling_factor),
                'gaussian_fwhm': int(self.gaussian_size),
                'convolution_method': str(self.convolution_method),
                'gaussian_kernel_size': int(self.gaussian_size),
                'include_moffat': bool(self.include_moffat),
                'elliptical_moffat': bool(self.elliptical_moffat),
            }

            with h5py.File(path, 'w') as f:
                dset = f.create_dataset("kwargs_options", data=json.dumps(kwargs_model))
                dset = f.create_dataset("kwargs_PSF", data=json.dumps(convert_numpy_array_to_list(kwargs)))
                dset = f.create_dataset("Norm", data=norm)

                if save_output_level > 1:
                    if data is not None:
                        dset = f.create_dataset("Data", data=data)
                    if sigma_2 is not None:
                        dset = f.create_dataset("Sigma2", data=sigma_2)
                    if masks is not None:
                        dset = f.create_dataset("Masks", data=masks)

                if save_output_level > 2:
                    narrow = self.get_narrow_psf(**kwargs, norm=True)
                    full = self.get_full_psf(**kwargs, norm=True, high_res=True)
                    background = self.get_background(kwargs['kwargs_background'])
                    analytic = self.get_moffat(kwargs['kwargs_moffat'], norm=True)

                    dset = f.create_dataset("Narrow PSF", data=narrow)
                    dset = f.create_dataset("Full PSF", data=full)
                    dset = f.create_dataset("Background", data=background)
                    dset = f.create_dataset("Analytic", data=analytic)

                if save_output_level > 3:
                    full_psfs = np.zeros((self.M, self.image_size, self.image_size))
                    residuals = np.zeros((self.M, self.image_size, self.image_size))
                    scaled_residuals = np.zeros((self.M, self.image_size, self.image_size))
                    for i in range(self.M):
                        estimated_full_psf = self.model(i, **kwargs)
                        dif = data[i, :, :] - estimated_full_psf
                        rr = jnp.abs(dif) / jnp.sqrt(sigma_2[i, :, :])
                        full_psfs[i, :, :] = estimated_full_psf
                        residuals[i, :, :] = dif
                        scaled_residuals[i, :, :] = rr

                    dset = f.create_dataset("Full PSF cube", data=full_psfs)
                    dset = f.create_dataset("Residuals cube", data=residuals)
                    dset = f.create_dataset("Scaled residuals cube", data=scaled_residuals)

        else:
            raise NotImplementedError(f'Unrecognized format {format}. Choose between pkl and hdf5.')

    def smart_guess(self, data, fixed_background=True, guess_method='barycenter', masks=None, offset_limit=None):
        """
        Returns an initial guess of the kwargs, given the input data.

        :param data: array of shape (nimage, npix, npix) containing the input data
        :param fixed_background: fixes the background to 0
        :type fixed_background: bool
        :param guess_method: Method to guess the position of the point sources. Choose between 'barycenter' and 'max'
        :type guess_method: str
        :param offset_limit: Upper and lower bounds for the center of the star in "big" pixel. Will be used in the kwargs_down/up['kwargs_gaussian']['x0'], kwargs_down/up['kwargs_gaussian']['y0'].
        :type offset_limit: float

        :return: kwargs containing an initial guess of the parameters
        """

        initial_fwhm = 3.  # defined in pixel, this is a good enough guess
        initial_beta = 2.
        initial_background = jnp.zeros((self.image_size_up ** 2))

        # Positions (initialisation at the center of gravity)
        x0_est = np.zeros(len(data))
        y0_est = np.zeros(len(data))
        
        # need a grid of pixels for the center of gravity:
        X, Y = jnp.indices(data[0].shape)
        # we'll recenter the positions in the loop: 
        # ( -1 because coordinates start at 0)
        centerpos = (self.image_size - 1) / 2.

        if guess_method == 'barycenter':
            # calculate center of gravity for each epoch:
            for i in range(len(data)):
                currentimage = data[i]
                if masks is not None:
                    currentimage *= masks[i]
                    
                # total weight of the image:
                partition = np.nansum(currentimage)
                # first moment: weight coordinates by image values
                x0 = np.nansum(X * currentimage) / partition
                y0 = np.nansum(Y * currentimage) / partition
                # x0 and y0 have their origin at top-left of image.
                x0_est[i] = y0 - centerpos
                y0_est[i] = x0 - centerpos  # x and y needs to be inverted

        elif guess_method == 'max':
            # Positions (initialisation to the brightest pixel)
            x0_est = np.zeros(len(data))
            y0_est = np.zeros(len(data))
            for i in range(len(data)):
                indices = np.where(data[i, :, :] == data[i, :, :].max())
                x0_est[i] = (indices[1] - self.image_size / 2.)
                y0_est[i] = (indices[0] - self.image_size / 2.) #x and y needs to be inverted

        elif guess_method == 'center':
            # Positions (initialisation to the brightest pixel)
            x0_est = np.zeros(len(data))
            y0_est = np.zeros(len(data))
        else :
            raise ValueError('Guess methods unknown. PLease choose between "max", "center" and "barycenter".')

        # Amplitude (use the total flux to scale the amplitude parameters)
        mean_flux = data.sum() / len(data)
        ratio = jnp.array([data[i].sum() / mean_flux for i in range(len(data))])
        initial_a = jnp.ones(len(data)) * ratio
        if self.elliptical_moffat:
            kwargs_moffat_guess = {'fwhm_x': initial_fwhm,'fwhm_y': initial_fwhm, 'phi': 0., 'beta': initial_beta, 'C': 1.}
            kwargs_moffat_up = {'fwhm_x': self.image_size, 'fwhm_y': self.image_size, 'phi': np.pi / 2., 'beta': 50.,
                                'C': jnp.inf}
            kwargs_moffat_down = {'fwhm_x': 2., 'fwhm_y': 2., 'phi': 0., 'beta': 0., 'C': 0.}
        else:
            kwargs_moffat_guess = {'fwhm': initial_fwhm, 'beta': initial_beta, 'C': 1.}
            kwargs_moffat_up = {'fwhm': self.image_size, 'beta': 50., 'C': jnp.inf}
            kwargs_moffat_down = {'fwhm': 2., 'beta': 0., 'C': 0.}

        flux_moffat = Downsample(self.get_moffat(kwargs_moffat_guess, norm=False), factor=self.upsampling_factor).sum()
        initial_C = float(mean_flux / flux_moffat)

        kwargs_moffat_guess['C']=initial_C
        kwargs_init = {
            'kwargs_moffat': kwargs_moffat_guess,
            'kwargs_gaussian': {'a': initial_a * initial_C, 'x0': x0_est, 'y0': y0_est},
            'kwargs_background': {'background': initial_background},
        }

        kwargs_fixed = {
            'kwargs_moffat': {},
            'kwargs_gaussian': {},
            # 'kwargs_background': [{}],
            'kwargs_background': {},
        }
        if fixed_background:
            kwargs_fixed['kwargs_background']['background'] = initial_background

        if offset_limit is None:
            offset_limit = self.image_size / 2.

        # Default value for boundaries
        kwargs_up = {
            'kwargs_moffat': kwargs_moffat_up,
            'kwargs_gaussian': {'a': list([jnp.inf for i in range(len(data))]),
                                'x0': list([offset_limit for i in range(len(data))]),
                                'y0': list([offset_limit for i in range(len(data))])
                                },
            'kwargs_background': {'background': list([jnp.inf for i in range(self.image_size_up ** 2)])},
        }

        kwargs_down = {
            'kwargs_moffat': kwargs_moffat_down,
            'kwargs_gaussian': {'a': list([0 for i in range(len(data))]),
                                'x0': list([-offset_limit for i in range(len(data))]),
                                'y0': list([-offset_limit for i in range(len(data))]),
                                },
            'kwargs_background': {'background': list([-jnp.inf for i in range(self.image_size_up ** 2)])},
        }

        return kwargs_init, kwargs_fixed, kwargs_up, kwargs_down


def load_PSF_model(input_file, format='hdf5'):
    """ Load PSF model class from hdf5 or pickle file"""

    if format == 'pkl':
        with open(input_file, 'rb') as f:
            model, kwargs, norm = pkl.load(f)
            data, sigma_2, masks = None, None, None

    elif format == 'hdf5':
        with h5py.File(input_file, 'r') as f:
            kwargs_model = json.loads(f['kwargs_options'][()])
            model = PSF(**kwargs_model)
            kwargs = json.loads(f['kwargs_PSF'][()])
            kwargs = convert_list_to_numpy_array(kwargs)
            norm = f['Norm'][()]

            if 'Data' in f.keys():
                data = f['Data'][()]
            else:
                print(f'No Data found in {input_file}')
                data = None
            if 'Sigma2' in f.keys():
                sigma_2 = f['Sigma2'][()]
            else:
                print(f'No Noise maps found in {input_file}')
                sigma_2 = None
            if 'Masks' in f.keys():
                masks = f['Masks'][()]
            else:
                print(f'No masks found in {input_file}')
                masks = None
    else:
        raise NotImplementedError(f'Unrecognized format {format}. Choose between pkl and hdf5.')
    return model, kwargs, norm, data, sigma_2, masks



from functools import partial

import jax.numpy as jnp
from jax import jit, vmap
from jax.lax import conv_general_dilated, conv_dimension_numbers
from jax.scipy.ndimage import map_coordinates
from jax.image import scale_and_translate

def scale_norms(n_scales):
    npix_dirac = 2 ** (n_scales + 2)
    dirac = jnp.diag((jnp.arange(npix_dirac) == int(npix_dirac / 2)).astype(float))
    wt_dirac = decompose(dirac, n_scales)
    norms = jnp.sqrt(jnp.sum(wt_dirac ** 2, axis=(1, 2,)))

    return norms

@partial(jit, static_argnums=(2))
def convolveSeparableDilated(image2D, kernel1D, dilation=1):
    """
    Convolves an image contained in ``image2D`` with the 1D kernel ``kernel1D``. The operation is
    blured2D = image2D * (kernel1D ∧ kernel1D), where ∧ is a wedge product, here a tensor product.

    :param kernel1D: 1D array to convolve the image with
    :param image2D: 2D array
    :param dilation:  makes the spacial extent of the kernel bigger. The default is 1.

    :return: 2D array

    """

    # Preparations
    image = jnp.expand_dims(image2D, (2,))
    # shape (Nx, Ny, 1) -- (N, W, C)
    # we treat the Nx as the batch number!! (because it is a 1D convolution
    # over the rows)
    kernel = jnp.expand_dims(kernel1D, (0, 2,))
    # here we have kernel shape ~(I,W,O)

    # so:
    # (Nbatch, Width, Channel) * (Inputdim, Widthkernel, Outputdim)
    #                                            -> (Nbatch, Width, Channel)
    # where Nbatch is our number of rows.
    dimension_numbers = ('NWC', 'IWO', 'NWC')
    dn = conv_dimension_numbers(image.shape,
                                kernel.shape,
                                dimension_numbers)
    # with these conv_general_dilated knows how to handle the different
    # axes:
    rowblur = conv_general_dilated(image, kernel,
                                   window_strides=(1,),
                                   padding='SAME',
                                   rhs_dilation=(dilation,),
                                   dimension_numbers=dn)

    # now we do the same for the columns, hence this time we have
    # (Height, Nbatch, Channel) * (Inputdim, Widthkernel, Outputdim)
    #                                            -> (Height, Nbatch, Channel)
    # where Nbatch is our number of columns.
    dimension_numbers = ('HNC', 'IHO', 'HNC')
    dn = conv_dimension_numbers(image.shape,
                                kernel.shape,
                                dimension_numbers)

    rowcolblur = conv_general_dilated(rowblur, kernel,
                                      window_strides=(1,),
                                      padding='SAME',
                                      rhs_dilation=(dilation,),
                                      dimension_numbers=dn)

    return rowcolblur[:, :, 0]

@partial(jit, static_argnums=(1,))
def decompose(image, n_scales):
    """Decomposes an image into a chosen wavelet basis."""
    # Validate input
    assert n_scales >= 0, "nscales must be a non-negative integer"
    if n_scales == 0:
        return image

    # Preparations
    image = jnp.copy(image)
    kernel = jnp.array([1, 4, 6, 4, 1]) / 16. #kernel for wavelet

    # Compute the first scale:
    c1 = convolveSeparableDilated(image, kernel)
    # Wavelet coefficients:
    w0 = (image - c1)
    result = jnp.expand_dims(w0, 0)
    cj = c1

    # Compute the remaining scales
    # at each scale, the kernel becomes larger ( a trou ) using the
    # dilation argument in the jax wrapper for convolution.
    for step in range(1, n_scales):
        cj1 = convolveSeparableDilated(cj, kernel, dilation=2 ** step)
        # wavelet coefficients
        wj = (cj - cj1)
        result = jnp.concatenate((result, jnp.expand_dims(wj, 0)), axis=0)
        cj = cj1

    # Append final coarse scale
    result = jnp.concatenate((result, jnp.expand_dims(cj, axis=0)), axis=0)
    return result


def get_l1_norm(image, W, n_scales):
    st = decompose(image, n_scales)[:-1]  # ignore coarsest scale
    st_weighted_l1 = jnp.sum(W * jnp.abs(st))  # other scales

    return st_weighted_l1

@jit
def reconstruct(coeffs):
    """Reconstructs an image from wavelet decomposition coefficients."""
    return jnp.sum(coeffs, axis=0)


@jit
def rotate_array(array, angle_deg):
    """ 
    rotates a 2D array by angle_deg degrees. We are using bilinear interpolation (higher orders not implemented in jax yet) 


    :param array: 2D array to be rotated
    :param angle_deg: float, rotation in degrees.
    :return: 2D array - a rotated version of the input array
    """
    # coordinate grid
    y, x = jnp.indices(array.shape)
    coords = jnp.array([x.ravel(), y.ravel()])
    # shift to center
    coords_shifted = coords - jnp.array([[array.shape[1] // 2], [array.shape[0] // 2]])

    # rotate the grid by the opposite rotation
    angle_rad = - jnp.radians(angle_deg)
    x, y = coords_shifted
    x_rot = x * jnp.cos(angle_rad) - y * jnp.sin(angle_rad)
    y_rot = x * jnp.sin(angle_rad) + y * jnp.cos(angle_rad)

    # shift back and clip
    x_rot += array.shape[1] // 2
    y_rot += array.shape[0] // 2

    # map
    rotated_array = map_coordinates(array, [y_rot, x_rot], order=1).reshape(array.shape)

    return rotated_array


@jit
def translate_array(array, dx, dy):
    """
    Using first order interpolation, translates the array by (dx, dy) pixels.
                                                                                                                
    :param array: 2D array to be translated
    :param dx: translation in the x direction (in pixels)
    :type dx: float
    :param dy: translation in the y direction (in pixels)
    :type dy: float
    :return: 2D array - A translated version of the input array
                                                                                                                
    """
    return scale_and_translate(image=array,
                               shape=array.shape,
                               spatial_dims=(0, 1),
                               scale=jnp.array((1., 1.)),
                               translation=jnp.array((dy, dx)),
                               method='bilinear')



import jax.numpy as jnp
import numpy as np

from starred.utils.parameters import Parameters

__all__ = ['ParametersPSF']


class ParametersPSF(Parameters):
    """
    Point Spread Function parameters class.

    """
    param_names_moffat = ['fwhm_x', 'fwhm_y', 'phi', 'beta', 'C']
    param_names_background = ['background']
    param_names_gaussian = ['a', 'x0', 'y0']

    def __init__(self, kwargs_init, kwargs_fixed, kwargs_up=None, kwargs_down=None):
        """
        :param kwargs_init: dictionary with information on the initial values of the parameters
        :param kwargs_fixed: dictionary containing the fixed parameters 
        :param kwargs_up: dictionary with information on the upper bounds of the parameters 
        :param kwargs_down: dictionary with information on the lower bounds of the parameters 

        """
        super(ParametersPSF, self).__init__(kwargs_init, kwargs_fixed, kwargs_up=kwargs_up,
                                            kwargs_down=kwargs_down)

        # we'll guess the number of sources, etc. based on the provided kwargs
        self.M = len(kwargs_init['kwargs_gaussian']['x0'])  # number of sources, one per provided image
        self.background_param_number = len(kwargs_init['kwargs_background']['background'])
        # ah, and ...since we're making a strong distinction between elliptical or standard moffat, we'll
        # remove the extra params if we're not dealing with an elliptical moffat. However, in the future
        # we should change the parametrization to remove the need for this distinction:
        # can always fix the params of the elliptical to have a circular one.
        # TODO
        if not ('phi' in kwargs_init['kwargs_moffat']):
            self.param_names_moffat = ['fwhm', 'beta', 'C']

        self._update_arrays()

    def args2kwargs(self, args):
        """Obtain a dictionary of keyword arguments from positional arguments."""
        i = 0
        kwargs_moffat, i = self._get_params(args, i, 'kwargs_moffat')
        kwargs_gaussian, i = self._get_params(args, i, 'kwargs_gaussian')
        kwargs_background, i = self._get_params(args, i, 'kwargs_background')
        # wrap-up
        kwargs = {'kwargs_moffat': kwargs_moffat, 'kwargs_gaussian': kwargs_gaussian,
                  'kwargs_background': kwargs_background}

        return kwargs

    def kwargs2args(self, kwargs):
        """Obtain an array of positional arguments from a dictionary of keyword arguments."""
        args = self._set_params(kwargs, 'kwargs_moffat')
        args += self._set_params(kwargs, 'kwargs_gaussian')
        args += self._set_params(kwargs, 'kwargs_background')
        return jnp.array(args)

    def get_param_names_for_model(self, kwargs_key):
        """Returns the names of the parameters according to the key provided."""
        if kwargs_key == 'kwargs_moffat':
            return self.param_names_moffat
        elif kwargs_key == 'kwargs_gaussian':
            return self.param_names_gaussian
        elif kwargs_key == 'kwargs_background':
            return self.param_names_background
        else:
            raise KeyError(f'`{kwargs_key}` is not in the kwargs')

    def _get_params(self, args, i, kwargs_key):
        """Getting the parameters."""
        kwargs = {}
        kwargs_fixed_k = self._kwargs_fixed[kwargs_key]
        param_names = self.get_param_names_for_model(kwargs_key)
        for name in param_names:
            if not name in kwargs_fixed_k.keys():
                if name == 'background':
                    num_param = self.background_param_number
                elif name == 'a':
                    num_param = self.M
                elif name == 'x0' or name == 'y0':
                    num_param = self.M
                else:
                    num_param = 1
                kwargs[name] = args[i:i + num_param]
                i += num_param
            else:
                kwargs[name] = kwargs_fixed_k[name]
                free_ind = self._kwargs_free_indices[kwargs_key][name]
                if len(free_ind) > 0:
                    num_param = len(free_ind)
                    kwargs[name] = kwargs[name].at[free_ind].set(args[i:i+ num_param])
                    i += num_param

        return kwargs, i

    def get_all_free_param_names(self, kwargs):
        args = self._param_names(kwargs, 'kwargs_moffat')
        args += self._param_names(kwargs, 'kwargs_gaussian')
        args += self._param_names(kwargs, 'kwargs_background')
        return args



from copy import deepcopy
from functools import partial
from scipy.interpolate import griddata

import jax.numpy as jnp
import numpy as np
from jax import vmap, jit
from jax.lax import scan, dynamic_slice, dynamic_update_slice
import dill as pkl
import os
import h5py
import json

from starred.utils.generic_utils import pad_and_convolve_fft, \
    pad_and_convolve, scipy_convolve, \
    fwhm2sigma, gaussian_function, \
    make_grid, Downsample, save_fits, save_npy, convert_numpy_array_to_list, convert_list_to_numpy_array
from starred.utils.jax_utils import translate_array, rotate_array
from starred.utils.light_profile_analytical import sersic_ellipse


class Deconv(object):
    """
    Image deconvolution class. The order of the params below matters for unpacking re-packing dictionaries.

    """

    def __init__(self, image_size, number_of_sources, scale,
                 upsampling_factor=2, epochs=1, psf=None,
                 gaussian_fwhm=2, N_sersic=0, convolution_method='fft'):
        """
        :param image_size: input image size in pixels
        :type image_size: int
        :param number_of_sources: amount of point sources 
        :type number_of_sources: int
        :param scale: scaling factor
        :type scale: float
        :param upsampling_factor: the proportion by which the sampling rate increases with respect to the input image
        :type upsampling_factor: int
        :param epochs: for a joint deconvolution, the number of epochs considered. For single deconvolution, ``epochs`` is 1
        :type epochs: int
        :param psf: Point Spread Function array from ``starred.psf.psf``, acting as deconvolution kernel here
        :param gaussian_fwhm: the Gaussian's FWHM in pixels
        :type gaussian_fwhm: int
        :param convolution_method: method to use to calculate the convolution : 'fft', 'scipy' (recommended), or 'lax'
        :type convolution_method: str

        """
        self.image_size = image_size

        # setter since vectorized functions used here depend on upsampling factor:
        self.setUpsamplingFactor(upsampling_factor)

        self.image_size_up = image_size * upsampling_factor
        self.epochs = epochs
        self.M = number_of_sources
        self.N_sersic = N_sersic
        self.gaussian_fwhm = gaussian_fwhm
        # let's carry the coordinates of our working grid (the upscaled one):
        self.x, self.y = make_grid(numPix=self.image_size_up, deltapix=1.)

        if psf is None:
            raise TypeError('Please provide the narrow PSF.')
        else:
            if not len(psf.shape) == 3:
                raise RuntimeError(
                    'We expect your psf to have shape (epochs, nx, ny). Your psf does not have the right number of dimensions, with shape: {psf.shape}')
            # psf = np.flip(psf, (1, 2))
            n_epochs, nx, ny = psf.shape
            assert n_epochs == self.epochs, "Please provide one PSF for each epoch"

            # for convolution_method = 'fft', image and kernel must have the same size
            if (nx != self.image_size_up or ny != self.image_size_up) and convolution_method == 'fft':
                print('Padding the PSFs to match the upsampled image size...')
                if (self.image_size_up - nx) % 2 or (self.image_size_up - ny) % 2:
                    raise RuntimeError(
                        'Upsampled image size and PSF size do not match. Please pad the PSF manually to match '
                        'the size of the upsampled image. Make sure it is properly centered. '
                        'PSF kernels should have an even number of pixels if your are using convolution method "fft".')
                else:
                    padx, pady = int((self.image_size_up - nx) / 2), int((self.image_size_up - ny) / 2)
                    psf = jnp.pad(psf, ((0, 0), (padx, padx), (pady, pady)), constant_values=0.)

            self.psf = psf.astype(jnp.float32)

        if convolution_method == 'fft':
            self._convolve = pad_and_convolve_fft
            if nx % 2 == 0:
                self.shift_gaussian = -0.5
            else:
                self.shift_gaussian = 0.
            self.shift_direction = 1
        elif convolution_method == 'lax':
            self._convolve = pad_and_convolve
            if nx % 2 == 0:
                self.shift_gaussian = -0.5
            else:
                self.shift_gaussian = 0.
            self.shift_direction = 1
        elif convolution_method == 'scipy':
            self._convolve = scipy_convolve
            if nx % 2 == 0:
                self.shift_gaussian = 0.5
            else:
                self.shift_gaussian = 0.
            self.shift_direction = 1
        else:
            raise NotImplementedError('Unknown convolution method: choose fft, scipy (recommended) or lax')

        self.convolution_method = convolution_method
        self.sigma = fwhm2sigma(self.gaussian_fwhm)

        self._model = self.modelstack if self.epochs > 5 else self.modelstack_forloop
        self.scale = scale

    def shifted_gaussians(self, c_x, c_y, a, source_ID=None):
        """
        Generates a 2D array with the point sources of the deconvolved image.

        :param c_x: 1D array containing the x positions of the point sources in unit of "big" pixel
        :param c_y: 1D array containing the y positions of the point sources in unit of "big" pixel
        :param a: 1D array containing the amplitude coefficients of each Gaussian representing a point source
        :return: 2D array

        """
        if source_ID is None:
            source_ID = range(self.M)
        return jnp.sum(jnp.array([gaussian_function(
            x=self.x, y=self.y,
            amp=a[i], sigma_x=self.sigma,
            sigma_y=self.sigma,
            center_x=(self._upsampling_factor * c_x[i] - self.shift_gaussian) * self.shift_direction,
            center_y=(self._upsampling_factor * c_y[i] - self.shift_gaussian) * self.shift_direction
        ).reshape(self.image_size_up, self.image_size_up) for i in range(self.M) if i in source_ID]), axis=0)

    def sersic_background(self, R_sersic=[1.], n_sersic=[1.], e1=[0.], e2=[0.], center_x=[0.], center_y=[0.], amp=[0.]):
        """
        Return a 1D vector containing all the pixels of the Sersic background.

        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius, in big pixel
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        """

        back = jnp.zeros(self.image_size_up ** 2)
        if self.N_sersic > 0:
            for i in range(self.N_sersic):
                back += sersic_ellipse(x=self.x, y=self.y, R_sersic=self._upsampling_factor * R_sersic[i],
                                       n_sersic=n_sersic[i], e1=e1[i], e2=e2[i],
                                       center_x=(
                                                        self._upsampling_factor * center_x[
                                                    i] - self.shift_gaussian) * self.shift_direction,
                                       center_y=(
                                                        self._upsampling_factor * center_y[
                                                    i] - self.shift_gaussian) * self.shift_direction,
                                       amp=amp[i])
        return back

    def _get_background(self, kwargs):
        """
        Return the pixelated background (it does not add the mean per epoch, just the model).

        """
        kwargs_background = kwargs['kwargs_background']
        kwargs_sersic = kwargs['kwargs_sersic']

        h = kwargs_background['h']
        sersic = self.sersic_background(**kwargs_sersic)

        return h + sersic

    @partial(jit, static_argnums=(0,))
    def one_epoch(self, psf, params, h):
        """
        Produces a 2D array corresponding to one epoch of the deconvolution.

        :param psf: 2D array representing the PSF at this epoch, usually upsampled
        :param params: 1D array with the parameters in the following order:

            - the x positions of the point sources (M params)
            - the y positions of the point sources (M params)
            - the amplitudes of the point sources (M params)
            - the x translation of the whole epoch (1 param)
            - the y translation of the whole epoch (1 param)
        :param h:
            a 1D array containing the unique background of the deconvolution, ``self.image_size_up**2`` params.

        :return: 2D array that represents the model for one epoch. This is, a collection of point sources plus background, convolved with the PSF and potentially translated in the x and y directions

        """

        # unpack:
        # (a bit messy, see the re-packing in self.model and the comment there)
        # TODO
        c_x = params[0:self.M]
        c_y = params[self.M:2 * self.M]
        a = params[2 * self.M:3 * self.M]
        dx = params[3 * self.M]
        dy = params[3 * self.M + 1]
        alpha = params[3 * self.M + 2]
        mean = params[3 * self.M + 3]

        # initially, just the background:
        f = h.reshape((self.image_size_up, self.image_size_up))

        # add the point sources at the deconvolved resolution:
        f += self.scale * self.shifted_gaussians(c_x, c_y, a)

        # rotate
        f = rotate_array(f, alpha)

        # shift to compensate small translations between epochs:
        f = translate_array(f, self._upsampling_factor * dx, self._upsampling_factor * dy)

        # convolve with the psf to match the resolution of the observation:
        f = self._convolve(f, psf).reshape(self.image_size_up,
                                           self.image_size_up)

        # now we can also add the constant background:
        f += mean
        # adding after translation avoids border effects 
        # (translation pads image with 0s, very noticeable when mean != 0)

        return f

    def getDeconvolved(self, kwargs, epoch=None):
        """
        This is a utility function to retrieve the deconvolved model at one 
        specific epoch. It will typically be called only once after the optimization process.

        :param kwargs: dictionary containing the parameters of the model
        :param epoch: number of epochs, if none, it will return the mean image
        :type epoch: int
        :return: tuple of two 2D arrays: (deconvolved image, background)

        """
        kwargs_analytic = kwargs['kwargs_analytic']
        kwargs_background = kwargs['kwargs_background']

        c_x = kwargs_analytic['c_x']
        c_y = kwargs_analytic['c_y']
        if epoch is None:
            a = self.mean_pts_amps(kwargs_analytic['a'])
            dx = 0.
            dy = 0.
            alpha = 0.
            mean = jnp.mean(kwargs_background['mean'])
        else:
            a = kwargs_analytic['a'][self.M * epoch:self.M * (epoch + 1)]
            dx = kwargs_analytic['dx'][epoch]
            dy = kwargs_analytic['dy'][epoch]
            alpha = kwargs_analytic['alpha'][epoch]
            mean = kwargs_background['mean'][epoch]

        h = self._get_background(kwargs).reshape(self.image_size_up, self.image_size_up)
        h = h.copy()
        f = h.copy()
        if self.M != 0:
            f += self.scale * self.shifted_gaussians(c_x, c_y, a)

        deconvolved = rotate_array(f, alpha)
        deconvolved = translate_array(deconvolved, self._upsampling_factor * dx,
                                      self._upsampling_factor * dy)  # TODO : + mean here ?

        background = rotate_array(h, alpha)
        background = translate_array(background, self._upsampling_factor * dx, self._upsampling_factor * dy) + mean

        return deconvolved, background

    def getPts_source(self, kwargs, epoch=None, source_ID=None):
        """
        This is a utility function to retrieve an image with only the point source components at one
        specific epoch.

        :param kwargs: dictionary containing the parameters of the model
        :param epoch: number of epochs. If none, it will return the mean image
        :type epoch: int
        :return: 2D array.
        """
        kwargs_analytic = kwargs['kwargs_analytic']
        kwargs_background = kwargs['kwargs_background']

        c_x = kwargs_analytic['c_x']
        c_y = kwargs_analytic['c_y']
        if epoch is None:
            a = self.mean_pts_amps(kwargs_analytic['a'])
            dx = 0.
            dy = 0.
            alpha = 0.
        else:
            a = kwargs_analytic['a'][self.M * epoch:self.M * (epoch + 1)]
            dx = kwargs_analytic['dx'][epoch]
            dy = kwargs_analytic['dy'][epoch]
            alpha = kwargs_analytic['alpha'][epoch]

        f = jnp.zeros_like(kwargs_background['h'].reshape((self.image_size_up, self.image_size_up)))

        if self.M != 0:
            f += self.scale * self.shifted_gaussians(c_x, c_y, a, source_ID=source_ID)
            f = rotate_array(f, alpha)
            f = translate_array(f, self._upsampling_factor * dx, self._upsampling_factor * dy)

        return f

    def setUpsamplingFactor(self, upsampling_factor):
        """
        Sets the upsampling factor.

        :param upsampling_factor: the proportion by which the sampling rate increases with respect to the input image
        :type upsampling_factor: int

        """
        self._upsampling_factor = upsampling_factor

        # down sample operator from utils: will need a vectorized version.
        self.DownsampleAllEpochs = vmap(lambda array: Downsample(array, upsampling_factor),
                                        in_axes=(0,))

    @partial(jit, static_argnums=(0,))
    def modelstack_scan(self, pars):
        # unused. keeping for a while just in case.
        # lax.scan function works as: carry, output = f(carry, input)
        # make f below: (we'll use a decoy for carry as we are not using it.)
        one_epoch = lambda pair: self.one_epoch(pair[0], pair[1])
        # pair is (psf, params)
        f = lambda carry, x: (None, one_epoch(x))
        # scan(f, init_carry, params)
        return scan(f, None, (self.psf, pars))[1]

    @partial(jit, static_argnums=(0,))
    def modelstack_forloop(self, pars, h):
        """
        Creates a model for all the epochs with shape (epoch, image_size_up, image_size_up) by looping over the epochs.
        This is slightly faster for a low number of epochs (~1 to 5) than using a vectorized version of ``self.one_epoch``.

        :param pars: stack of arrays of parameters for one epoch. Shape (epoch, N), where N is the number of parameters
                     per epoch (see the docstring of ``self.one_epoch``)
        :param h: 1D array containing the pixelated background. Separated from `pars` because does not vary between epochs.

        :return: 2D array of shape (epoch, image_size_up, image_size_up)

        """
        return jnp.array([self.one_epoch(self.psf[i], pars[i], h) for i in range(self.epochs)])

    @partial(jit, static_argnums=(0,))
    def modelstack(self, pars, h):
        """
        Creates a model for all the epochs with shape (epoch, image_size_up, image_size_up) by vectorizing ``self.one_epoch``.
        Much faster for a large number of epochs (> 10), regime where the ``self.modelstack_forloop`` version cannot be jit'ed anymore.

        :param pars: stack of arrays of parameters for one epoch. Shape (epoch, N), where N is the number of parameters per epoch (see the docstring of ``self.one_epoch``)
        :param h: 1D array containing the pixelated background. Separated from `pars` because does not vary between epochs.

        :return: 2D array of shape (epoch, image_size_up, image_size_up)

        """

        # vmap magic to vectorize our one_epoch method:
        h_fixed = partial(self.one_epoch, h=h)  # with the background fixed in the function, since always the same.
        broadcast = vmap(h_fixed, in_axes=(0, 0))
        # here we specify `in_axes`: vectorize over the first axis of the
        # array of PSFs, and over the first axis of the array of parameters.
        # by the way:  
        # - array of PSFs: stack of 2D slices
        # - array of parameters: stack of 1D slices, each containing the parameters
        #                        for one epoch.

        return broadcast(self.psf, pars)

    @partial(jit, static_argnums=(0,))
    def model(self, kwargs):
        """
        Creates the 2D deconvolution model image.
        
        :param kwargs: dictionary containing the parameters of the model.
        :return: 3D array - A stack of 2D images (one layer per epoch)

        """

        # part 1: unpack the keywords and put them in a giant 2D array:
        an = deepcopy(kwargs['kwargs_analytic'])
        back = self._get_background(kwargs)
        mean = kwargs['kwargs_background']['mean']

        # a bit messy here, we manually unpack each keyword.
        # the ParameterDeconv class should be taking care of this,
        # but we don't have access to it in this class in the way
        # things are set up right now.
        # TODO 

        # vectorized unpacking for insane performance bro
        row1 = jnp.tile(jnp.hstack([an[key] for key in ['c_x', 'c_y']]), (self.epochs, 1))
        i_values = jnp.arange(self.epochs)[:, None]
        multiplier = i_values * self.M
        indices = jnp.arange(self.M) + multiplier  # generates a 2D array of indices
        row2 = an['a'][indices.flatten()].reshape(self.epochs, self.M)  # Flatten indices and reshape results

        row3 = an['dx'].reshape(self.epochs, 1)
        row4 = an['dy'].reshape(self.epochs, 1)
        row5 = an['alpha'].reshape(self.epochs, 1)
        epochmean = mean[i_values]
        # nice 2D array, containing Nepochs 1D arrays of all params.
        args = jnp.hstack([row1, row2, row3, row4, row5, epochmean])

        # part 2, pass the giant 2D array to the model function:
        modelupscale = self._model(args, back)
        # this generates a model on the upscaled grid.

        # Now we down sample. Each new pixel is the average of all the
        # small pixels that composed it before.        
        model = self.DownsampleAllEpochs(modelupscale)

        # to conserve the surface brightness:
        model = model * ((self.image_size_up ** 2) / (self.image_size ** 2))

        return model

    def dump(self, path, kwargs, data=None, sigma_2=None, save_output_level=4, format='hdf5'):
        """
        Stores the model in a given file in pickle or hdf5 format (recommended).

        :param path: Filename of the output.
        :param kwargs: Dictionary containing the fitted value of the model
        :param data: (Nepoch x image_size x image_size) array containing the data
        :param sigma_2: (Nepoch x image_size x image_size) array containing the noise maps
        :param save_output_level: Int. Level of output product to save: 1-just the parameters of the model, 2- add the input data, 3- add the output products (background, and pts source channel) 4- add the output products for every image.
        """

        if format == 'pkl':
            with open(path, 'wb') as f:
                pkl.dump([self, kwargs, data, sigma_2], f, protocol=pkl.HIGHEST_PROTOCOL)
        elif format == 'hdf5':
            kwargs_model = {
                'image_size': int(self.image_size),
                'number_of_sources': int(self.M),
                'scale': float(self.scale),
                'upsampling_factor': int(self._upsampling_factor),
                'epochs': int(self.epochs),
                'psf': np.asarray(self.psf).tolist(),
                'gaussian_fwhm': int(self.gaussian_fwhm),
                'N_sersic': int(self.N_sersic),
                'convolution_method': str(self.convolution_method),
            }
            with h5py.File(path, 'w') as f:
                dset = f.create_dataset("kwargs_options", data=json.dumps(kwargs_model))
                dset = f.create_dataset("kwargs_Deconv", data=json.dumps(convert_numpy_array_to_list(kwargs)))

                if save_output_level > 1:
                    if data is not None:
                        dset = f.create_dataset("Data", data=data)
                    if sigma_2 is not None:
                        dset = f.create_dataset("Sigma2", data=sigma_2)

                if save_output_level > 2:
                    h = self._get_background(kwargs).reshape(self.image_size_up, self.image_size_up)
                    pts_source_mean = self.getPts_source(kwargs, epoch=None)
                    dset = f.create_dataset("Background", data=h)
                    dset = f.create_dataset("Point source mean", data=pts_source_mean)

                if save_output_level > 4:
                    models = np.zeros((self.epochs, self.image_size, self.image_size))
                    residuals = np.zeros((self.M, self.image_size, self.image_size))
                    deconvolutions = np.zeros((self.M, self.image_size_up, self.image_size_up))
                    for i in range(self.epochs):
                        output = self.model(kwargs)[i]
                        deconv, h = self.getDeconvolved(kwargs, i)
                        data = data[i, :, :]
                        dif = data - output

                        models[i, :, :] = output  # low resolution model
                        residuals[i, :, :] = dif
                        deconv[i, :, :] = deconv  # high resolution model

                    dset = f.create_dataset("Models cube", data=models)
                    dset = f.create_dataset("Residuals cube", data=residuals)
                    dset = f.create_dataset("deconvolution cube", data=deconvolutions)

        else:
            raise NotImplementedError(f'Unrecognized format {format}. Choose between pkl and hdf5.')

    def export(self, output_folder, kwargs_final, data, sigma_2, epoch=None, norm=1, format='fits', header=None):
        """
        Saves all the output files.

        :param output_folder: path to the output folder
        :type output_folder: str
        :param kwargs_final: dictionary containing all keyword arguments
        :param data: array containing the images
        :param sigma_2: array containing the noise maps
        :param epoch: integer or array or list containing the indices of the epoch to export. None, for exporting all epochs.
        :param norm: normalisation, to have the outputs in the correct unit
        :param format: output format. Choose between ``npy`` or ``fits``
        :type format: str
        :param header: HDU header, as return by astropy.io.fits (optional). This is used to propagate the WCS coordinate to the outputfiles
        """

        if format == 'fits':
            save_fct = save_fits
        elif format == 'npy':
            save_fct = save_npy
        else:
            raise NotImplementedError(f'Format {format} unknown.')

        if epoch is None:
            print('Exporting all epochs')
            epoch_arr = np.arange(self.epochs)
        elif not hasattr(epoch, "__len__"):
            epoch_arr = np.array([epoch])
        else:
            epoch_arr = np.array(epoch)

        high_res_header = deepcopy(header)
        if header is not None:
            # change the WCS coordinate for the high resolution output images
            for k in ['CDELT1', 'CDELT2']:
                if k in high_res_header.keys():
                    high_res_header[k] = header[k] / self._upsampling_factor
            for k in ['CRPIX1', 'CRPIX2']:
                if k in high_res_header.keys():
                    high_res_header[k] = header[k] * self._upsampling_factor

        for i, e in enumerate(epoch_arr):
            output = self.model(kwargs_final)[e] * norm
            deconv, h = self.getDeconvolved(kwargs_final, e)
            deconv *= norm
            h *= norm
            data_show = data[e, :, :] * norm
            dif = data_show - output
            rr = dif / np.sqrt(sigma_2[e, :, :] * norm ** 2)

            save_fct(data_show, os.path.join(output_folder, 'data_{0:05d}'.format(i)), header=header)
            save_fct(output, os.path.join(output_folder, 'model_{0:05d}'.format(i)), header=header)
            save_fct(dif, os.path.join(output_folder, 'residuals_{0:05d}'.format(i)), header=header)
            save_fct(rr, os.path.join(output_folder, 'scaled_residuals_{0:05d}'.format(i)), header=header)
            save_fct(deconv, os.path.join(output_folder, 'deconvolution_{0:05d}'.format(i)), header=high_res_header)

        h = self._get_background(kwargs_final).reshape(self.image_size_up, self.image_size_up)
        save_fct(h, os.path.join(output_folder, 'background_model'), header=high_res_header)

    def flux_at_epoch(self, kwargs, epoch=0):
        """
        Return an array containing the flux of the point sources at epoch `epoch`.

        :param kwargs: dictionary containing all keyword arguments
        :param epoch: index of the epoch
        """
        a = kwargs['kwargs_analytic']['a'][self.M * epoch:self.M * (epoch + 1)] * self.scale

        return a

    def get_mask_pts_source(self, kwargs, mask_size=3, nan_mask=False):
        """
        Return an array containing a mask of size 'mask_size' around the point sources. Ouput array is at the
        sub-sampled resolution.

        :param kwargs: dictionary containing all keyword arguments
        :param mask_size: size of the mask around the point sources. In "small" pixel.
        :param nan_mask: If true, return a mask where 0. have been replaced by nan

        """

        x0 = jnp.floor(kwargs['kwargs_analytic'][
                           'c_x'] * self._upsampling_factor + self.image_size_up / 2. - mask_size / 2.).astype(
            int)
        y0 = jnp.floor(kwargs['kwargs_analytic'][
                           'c_y'] * self._upsampling_factor + self.image_size_up / 2. - mask_size / 2.).astype(
            int)

        masks_pts_source = jnp.ones((self.image_size_up, self.image_size_up))

        for i in range(self.M):
            if nan_mask:
                replacement = jnp.nan * jnp.ones((mask_size, mask_size))
                masks_pts_source = dynamic_update_slice(masks_pts_source, replacement,
                                                        (y0[i], x0[i]))  # inverted indices
            else:
                replacement = jnp.zeros((mask_size, mask_size))
                masks_pts_source = dynamic_update_slice(masks_pts_source, replacement,
                                                        (y0[i], x0[i]))  # inverted indices

        return masks_pts_source

    def mean_pts_amps(self, a):
        """
        Return the mean amplitude per points source, over all epoch. Return a 1D vector of size self.M.

        :param a: amplitude vector, correspond to kwargs['kwargs_analytic']['a']
        """
        mean_amp = a.reshape(self.epochs, self.M)
        return jnp.mean(mean_amp, axis=0)

    def background_interpolate(self, kwargs, interpolation_order=0, inner_mask_size=3, outer_mask_size=7):
        """
        Return an interpolated version of the background, interpolated below the point sources.

        :param kwargs: dictionary containing all keyword arguments
        :param interpolation_order: order of the interpolation. For now, only 0 is available. It will take the mean of the pixel in the outer mask.
        :param inner_mask_size: size of the region to interpolate
        :param outer_mask_size: size of the outer mask used for the interpolation

        """
        assert outer_mask_size > inner_mask_size, "Outer mask should be bigger than the inner mask!"

        interp_back = deepcopy(kwargs['kwargs_background']['h']).reshape(self.image_size_up,
                                                                         self.image_size_up)

        x0_inner = jnp.floor(kwargs['kwargs_analytic'][
                                 'c_x'] * self._upsampling_factor + self.image_size_up / 2. - inner_mask_size / 2.).astype(
            int)
        y0_inner = jnp.floor(kwargs['kwargs_analytic'][
                                 'c_y'] * self._upsampling_factor + self.image_size_up / 2. - inner_mask_size / 2.).astype(
            int)
        x0_outer = jnp.floor(kwargs['kwargs_analytic'][
                                 'c_x'] * self._upsampling_factor + self.image_size_up / 2. - outer_mask_size / 2.).astype(
            int)
        y0_outer = jnp.floor(kwargs['kwargs_analytic'][
                                 'c_y'] * self._upsampling_factor + self.image_size_up / 2. - outer_mask_size / 2.).astype(
            int)

        if interpolation_order == 0:
            for i in range(self.M):
                # get the outer mask
                outerslice = dynamic_slice(interp_back, (y0_outer[i], x0_outer[i]), (outer_mask_size, outer_mask_size))
                replacement = jnp.nan * jnp.ones((outer_mask_size - 2, outer_mask_size - 2))
                masked_outerslice = dynamic_update_slice(outerslice, replacement,
                                                         (1, 1))

                # compute the mean in the outer ring around pts sources
                mean_outer_ring = jnp.nanmean(masked_outerslice)

                # replace the inner mask by the interpolated value
                replacement = jnp.ones((inner_mask_size, inner_mask_size)) * mean_outer_ring
                interp_back = dynamic_update_slice(interp_back, replacement,
                                                   (y0_inner[i], x0_inner[i]))  # inverted indices
        else:
            raise NotImplementedError('Higher order interpolation not implemented yet in JAX.')

        return interp_back, masked_outerslice, replacement




def setup_model(data, sigma_2, s, xs, ys, subsampling_factor, initial_a=None, astrometric_bound=None,
                dithering_bound=None, convolution_method='scipy', N_sersic=0):
    """
    Utility setting up a deconvolution model. The returned dictionaries of
    parameters can later be adjusted by the user. 

    :param data: 3D array containing the images, one per epoch. shape (epochs, im_size, im_size)
    :param sigma_2: 3D array containing the noisemaps, one per epoch. shape (epochs, im_size, im_size)
    :param s: 3D array containing the narrow PSFs, one per epoch. shape (epochs, im_size_up, im_size_up) where im_size_up needs be a multiple of im_size.
    :param xs: 1D array or list containing the x positions of the point sources. For M point sources, len(xs) is M. In units of big pixel.
    :param ys: 1D array or list containing the y positions of the point sources. For M point sources, len(ys) is M. In units of big pixel.
    :param initial_a: list containing the amplitudes of the point sources. For M point sources and N epochs, len(initial_a) is M*N. If none, amplitudes are applied scaled by the maax of each epoch.
    :param subsampling_factor: integer, ratio of the size of the data pixels to that of the PSF pixels.
    :param astrometric_bound: integer, maximum shift of the point sources, relative to the initial position. None: no upper nor lower limit
    :param dithering_bound: integer, maximum shift from image to image. None: no upper nor lower limit
    :param convolution_method: 'scipy', 'fft', or 'lax'. To be passed to the Deconv class. Recommended : 'scipy'.
    :param N_sersic: Number of sersic profile to add to the background in addition to the pixelated grid (Default: 0)

    :return: a tuple: (a starred.deconvolution.Deconv instance,
                       a dictionary with the initial values of the model parameters,
                       a dictionary with the upper boundaries of the model parameters,
                       a dictionary with the lower boundaries of the model parameters,
                       a dictionary with the fixed parameters of the model)
    """
    # xs and ys, let's make them numpy arrays cuz it's easier
    xs = np.array(xs)
    ys = np.array(ys)
    M = xs.size  # number of point sources
    epochs = data.shape[0]
    N_sersic = int(N_sersic)

    if initial_a is None:
        initial_a = 6 * np.array([np.nanmax(data[i, :, :]) for i in range(epochs) for j in range(M)])

    if not xs.size == ys.size:
        message = "Your xs or ys (refering to amplitudes and positions of your point sources)"
        message += " arguments need be the of the same length!\n"
        message += f"But we have size(xs)={xs.size}, size(ys)={ys.size}"
        raise RuntimeError(message)

    if not int(len(initial_a) / epochs) == xs.size:
        message = "Your initial amplitudes does not match the number of point sources and epochs."
        message += f" initial amplitudes must have length N_point_source * N_epochs = {M * epochs}.\n"
        message += f"But we have size(initial_a)={len(initial_a)}"
        raise RuntimeError(message)

    if not sigma_2.shape == data.shape:
        message = "The shape of your data and noisemaps is not what we expect."
        message += " They need be of shape ~ (epochs, some nx, some ny).\n"
        message += f"But we have shape(data)={data.shape}, shape(sigma_2)={sigma_2.shape}."
        raise RuntimeError(message)

    if not epochs == sigma_2.shape[0] == s.shape[0]:
        message = "The number of epochs in your data, noisemaps or PSFs do not match\n"
        message += f"We have shape(data)={data.shape}, shape(sigma_2)={sigma_2.shape}, shape(s)={s.shape}\n"
        message += "We expected the first dimension of each (# of epochs) to be the same."
        raise RuntimeError(message)

    im_size = data[0].shape[0]
    im_size_up = im_size * subsampling_factor
    if astrometric_bound is None:
        astrometric_bound = im_size / 2.
    if dithering_bound is None:
        dithering_bound = im_size / 2.

    # Parameter initialization
    initial_c_x = xs
    initial_c_y = ys
    # intensity per point:
    scale = np.nanmax(data)
    initial_a /= scale
    # initial background:
    initial_h = np.zeros((im_size_up ** 2))
    # dictionary containing the parameters of deconvolution.
    # (The translations dx, dy are set to zero for the first epoch.
    # Thus we only initialize (epochs-1) of them.)
    kwargs_init = {
        'kwargs_analytic': {
            'c_x': initial_c_x,  # point sources positions
            'c_y': initial_c_y,
            'dx': np.ravel([0. for _ in range(epochs)]),  # translation per epoch
            'dy': np.ravel([0. for _ in range(epochs)]),
            'alpha': np.ravel([0. for _ in range(epochs)]),  # rotation per epoch
            'a': initial_a},  # amplitudes of point sources
        'kwargs_background': {'h': initial_h,  # background
                              'mean': np.ravel([0. for _ in range(epochs)])},
        # additive constant for background per epoch
        'kwargs_sersic': {'amp': [0. for _ in range(N_sersic)],
                          'R_sersic': [1. for _ in range(N_sersic)],
                          'n_sersic': [1. for _ in range(N_sersic)],
                          'center_x': [0. for _ in range(N_sersic)],
                          'center_y': [0. for _ in range(N_sersic)],
                          'e1': [0. for _ in range(N_sersic)],
                          'e2': [0. for _ in range(N_sersic)],
                          }
    }
    # same as above, providing fixed parameters:
    kwargs_fixed = {
        'kwargs_analytic': {'alpha': kwargs_init['kwargs_analytic']['alpha']},
        'kwargs_background': {},
        'kwargs_sersic': {},
    }
    if epochs == 1:
        # we fix image translation
        kwargs_fixed['kwargs_analytic']['dx'] = [0.]
        kwargs_fixed['kwargs_analytic']['dy'] = [0.]

    # boundaries.
    kwargs_up = {
        'kwargs_analytic': {'c_x': list(initial_c_x + astrometric_bound),
                            'c_y': list(initial_c_y + astrometric_bound),
                            'dx': [dithering_bound for _ in range(epochs)],
                            'dy': [dithering_bound for _ in range(epochs)],
                            'alpha': [np.pi for _ in range(epochs)],
                            'a': list([np.inf for i in range(M * epochs)])
                            },
        'kwargs_background': {'h': list([np.inf for i in range(0, im_size_up ** 2)]),
                              'mean': [np.inf for _ in range(epochs)]
                              },
        'kwargs_sersic': {'amp': [np.inf for _ in range(N_sersic)],
                          'R_sersic': [100. for _ in range(N_sersic)],
                          'n_sersic': [8. for _ in range(N_sersic)],
                          'center_x': [im_size_up for _ in range(N_sersic)],
                          'center_y': [im_size_up for _ in range(N_sersic)],
                          'e1': [0.5 for _ in range(N_sersic)],
                          'e2': [0.5 for _ in range(N_sersic)],
                          },
    }
    kwargs_down = {
        'kwargs_analytic': {'c_x': list(initial_c_x - astrometric_bound),
                            'c_y': list(initial_c_y - astrometric_bound),
                            'dx': [-dithering_bound for _ in range(epochs)],
                            'dy': [-dithering_bound for _ in range(epochs)],
                            'alpha': [-np.pi for _ in range(epochs)],
                            'a': list([0 for i in range(M * epochs)])},
        'kwargs_background': {'h': list([-np.inf for i in range(0, im_size_up ** 2)]),
                              'mean': [-np.inf for _ in range(epochs)]
                              },
        'kwargs_sersic': {'amp': [0. for _ in range(N_sersic)],
                          'R_sersic': [0. for _ in range(N_sersic)],
                          'n_sersic': [0.5 for _ in range(N_sersic)],
                          'center_x': [-im_size_up for _ in range(N_sersic)],
                          'center_y': [-im_size_up for _ in range(N_sersic)],
                          'e1': [-0.5 for _ in range(N_sersic)],
                          'e2': [-0.5 for _ in range(N_sersic)],
                          },
    }
    # Initializing the model
    model = Deconv(image_size=im_size,
                   number_of_sources=M,
                   scale=scale,
                   upsampling_factor=subsampling_factor,
                   epochs=epochs,
                   psf=s,
                   convolution_method=convolution_method,
                   N_sersic=N_sersic)

    # returning model and kwargs.    
    return model, kwargs_init, kwargs_up, kwargs_down, kwargs_fixed


def load_Deconv_model(input_file, format='hdf5'):
    """ Load Deconv model class from hdf5 or pickle file"""

    if format == 'pkl':
        with open(input_file, 'rb') as f:
            model, kwargs, data, sigma_2 = pkl.load(f)

    elif format == 'hdf5':
        with h5py.File(input_file, 'r') as f:
            kwargs_model = json.loads(f['kwargs_options'][()])
            kwargs_model['psf'] = np.asarray(kwargs_model['psf'])  # convert back the list to np.array
            model = Deconv(**kwargs_model)
            kwargs = json.loads(f['kwargs_Deconv'][()])
            kwargs = convert_list_to_numpy_array(kwargs)

            if 'Data' in f.keys():
                data = f['Data'][()]
            else:
                print(f'No Data found in {input_file}')
                data = None
            if 'Sigma2' in f.keys():
                sigma_2 = f['Sigma2'][()]
            else:
                print(f'No Noise maps found in {input_file}')
                sigma_2 = None

    else:
        raise NotImplementedError(f'Unrecognized format {format}. Choose between pkl and hdf5.')

    return model, kwargs, data, sigma_2



import warnings

from starred.deconvolution.loss import Loss as deconvLoss
from starred.psf.loss import Loss as psfLoss
from starred.deconvolution.parameters import ParametersDeconv
from starred.utils.noise_utils import propagate_noise
from starred.optim.optimization import Optimizer
from starred.plots import plot_function as pltf
from starred.psf.parameters import ParametersPSF

from copy import deepcopy
import numpy as np


def multi_steps_deconvolution(data, model, parameters, sigma_2, s, subsampling_factor,
                              fitting_sequence=[['background'], ['pts-source-astrometry'], []],
                              optim_list=['l-bfgs-b', 'adabelief', 'adabelief'], kwargs_optim_list=None,
                              lambda_scales=1., lambda_hf=1000., lambda_positivity_bkg=100, lambda_positivity_ps=100,
                              lambda_pts_source=0., prior_list=None, regularize_full_model=False,
                              update_lambda_pts_source=False,
                              adjust_sky=False, noise_propagation='MC', regularization_terms='l1_starlet',
                              verbose=True):
    """
    A high-level function to run several time the optimisation algorithms and ensure to find the optimal solution.

    :param data: 3D array containing the images, one per epoch. shape (epochs, im_size, im_size)
    :param model: Deconv class, deconvolution model
    :param parameters: ParametersDeconv class
    :param sigma_2: 3D array containing the noisemaps, one per epoch. shape (epochs, im_size, im_size)
    :param s: 3D array containing the narrow PSFs, one per epoch. shape (epochs, im_size_up, im_size_up) where im_size_up needs be a multiple of im_size.
    :param subsampling_factor: integer, ratio of the size of the data pixels to that of the PSF pixels.
    :param fitting_sequence: list, List of lists, containing the element of the model to keep fixed. Example : [['pts-source-astrometry','pts-source-photometry','background'],['pts-source-astrometry','pts-source-photometry'], ...]
    :param optim_list: List of optimiser. Recommended if background is kept constant : 'l-bfgs-b', 'adabelief' otherwise.
    :param kwargs_optim_list: List of dictionary, containing the setting for the different optimiser.
    :param lambda_scales: Lagrange parameter that weights intermediate scales in the transformed domain
    :param lambda_hf: Lagrange parameter weighting the highest frequency scale
    :param lambda_pts_source: Lagrange parameter regularising the pts source channel
    :param lambda_positivity_bkg: Lagrange parameter weighting the positivity of the background. 0 means no positivity constraint.
    :param lambda_positivity_ps: Lagrange parameter weighting the positivity of the point sources. 0 means no positivity constraint.
    :param regularize_full_model: option to regularise just the background (False) or background + point source channel (True)
    :param update_lambda_pts_source : bool, option to refine the points channel regularisation strenght at every iteration
    :param prior_list: list of Prior object, Gaussian prior on parameters to be applied at each step
    :param adjust_sky: bool, True if you want to fit some sky subtraction
    :param noise_propagation: 'MC' or 'SLIT', method to compute the noise propagation in wavelet space. Default: 'MC'.
    :param verbose: bool. Verbosity.

    Return model, parameters, loss, kwargs_partial_list, fig_list, LogL_list, loss_history_list
    """
    # Check the sequence
    assert len(fitting_sequence) == len(optim_list), "Fitting sequence and optimiser list have different lenght !"
    if kwargs_optim_list is not None:
        assert len(fitting_sequence) == len(
            kwargs_optim_list), "Fitting sequence and kwargs optimiser list have different lenght !"
    else:
        warnings.warn('No optimiser kwargs provided. Default configuration is used.')
        kwargs_optim_list = [{} for _ in range(len(fitting_sequence))]

    if prior_list is None:
        prior_list = [None for _ in range(len(fitting_sequence))]
    else:
        assert len(fitting_sequence) == len(prior_list), "Fitting sequence and prior list have different lenght !"

    # setup the model
    kwargs_init, kwargs_fixed_default, kwargs_up, kwargs_down = deepcopy(parameters._kwargs_init), deepcopy(
        parameters._kwargs_fixed), \
        deepcopy(parameters._kwargs_up), deepcopy(parameters._kwargs_down)
    if not adjust_sky:
        kwargs_fixed_default['kwargs_background']['mean'] = np.zeros(model.epochs)

    kwargs_partial_list = [kwargs_init]
    fig_list = []
    loss_history_list = []
    loss_list = []

    W = propagate_noise(model, np.sqrt(sigma_2), kwargs_init, wavelet_type_list=['starlet'],
                        method=noise_propagation,
                        num_samples=500, seed=1, likelihood_type='chi2', verbose=False,
                        upsampling_factor=subsampling_factor)[0]

    for i, steps in enumerate(fitting_sequence):
        kwargs_fixed = deepcopy(kwargs_fixed_default)
        background_free = True
        for fixed_feature in steps:
            if fixed_feature == 'pts-source-astrometry':
                kwargs_fixed['kwargs_analytic']['c_x'] = kwargs_partial_list[i]['kwargs_analytic']['c_x']
                kwargs_fixed['kwargs_analytic']['c_y'] = kwargs_partial_list[i]['kwargs_analytic']['c_y']
            elif fixed_feature == 'pts-source-photometry':
                kwargs_fixed['kwargs_analytic']['a'] = kwargs_partial_list[i]['kwargs_analytic']['a']
            elif fixed_feature == 'background':
                background_free = False
                kwargs_fixed['kwargs_background']['h'] = kwargs_partial_list[i]['kwargs_background']['h']
            elif fixed_feature == 'sersic':
                for k in kwargs_partial_list[i]['kwargs_sersic'].keys():
                    kwargs_fixed['kwargs_sersic'][k] = kwargs_partial_list[i]['kwargs_sersic'][k]
            else:
                raise ValueError(
                    'Steps is not defined. Choose between "pts-source-astrometry", "pts-source-photometry", "background" or "sersic".')

        # need to recompile the parameter class, since we have changed the number of free parameters
        parameters = ParametersDeconv(kwargs_partial_list[i], kwargs_fixed, kwargs_up=kwargs_up,
                                      kwargs_down=kwargs_down)

        # for speed-up we turn of the regularisation to avoid the starlet decomposition, if the background is fixed
        if background_free:
            lambda_scales_eff, lambda_hf_eff, lambda_positivity_bkg_eff = deepcopy(lambda_scales), deepcopy(
                lambda_hf), deepcopy(lambda_positivity_bkg)
        else:
            lambda_scales_eff, lambda_hf_eff, lambda_positivity_bkg_eff = 0., 0., 0.

        # refine the regularisation of the point source channel correction from previous iteration
        if i > 0 and update_lambda_pts_source and lambda_pts_source != 0:
            lambda_pts_prev = lambda_pts_source_eff
            lambda_pts_source_eff = loss.update_lambda_pts_source(kwargs_partial_list[i])
            if verbose:
                print(f'Updating lambda_pts_source from {lambda_pts_prev} to {lambda_pts_source_eff}')
        else:
            lambda_pts_source_eff = lambda_pts_source

            # run the optimisation :
        print('Step %i, fixing :' % (i + 1), steps)
        loss = deconvLoss(data, model, parameters, sigma_2, regularization_terms=regularization_terms,
                          regularization_strength_scales=lambda_scales_eff,
                          regularization_strength_hf=lambda_hf_eff,
                          regularization_strength_positivity=lambda_positivity_bkg_eff,
                          regularization_strength_positivity_ps=lambda_positivity_ps,
                          regularization_strength_pts_source=lambda_pts_source_eff,
                          regularize_full_model=regularize_full_model,
                          prior=prior_list[i], W=W, )

        optim = Optimizer(loss, parameters, method=optim_list[i])
        best_fit, loss_best_fit, extra_fields, runtime = optim.minimize(**kwargs_optim_list[i])

        # Saving partial results
        kwargs_partial_steps = deepcopy(parameters.best_fit_values(as_kwargs=True))
        fig_list.append(
            pltf.plot_deconvolution(model, data, sigma_2, s, kwargs_partial_steps, epoch=0, units='e-', cut_dict=None))
        loss_history_list.append(extra_fields['loss_history'])
        loss_list.append(loss_best_fit)

        kwargs_partial_list.append(deepcopy(kwargs_partial_steps))
        if verbose:
            print('Step %i/%i took %2.f seconds' % (i + 1, len(fitting_sequence), runtime))
            print('Kwargs partial at step %i/%i' % (i + 1, len(fitting_sequence)), kwargs_partial_steps)
            print('Loss : ', loss_best_fit)
            print('Overall Reduced Chi2 : ',
                  -2 * loss._log_likelihood_chi2(kwargs_partial_steps) / (model.image_size ** 2))

    return model, parameters, loss, kwargs_partial_list, fig_list, loss_list, loss_history_list



import time
import jax
import jax.numpy as jnp
import numpy as np

import emcee
from mclmc.sampler import Sampler as MCHMCSampler
from starred.optim.inference_base import InferenceBase

SUPPORTED_SAMPLER = ['emcee', 'mchmc']


class Sampler(InferenceBase):
    def __init__(self, loss_class, param_class, sampler='emcee'):
        super().__init__(loss_class, param_class)

        if sampler not in SUPPORTED_SAMPLER:
            raise NotImplementedError(f'Sampler {sampler} is not supported yet.')
        else:
            self.sampler = sampler

        self.grad_nlogp = self.value_and_gradient
        self.d = len(self._param.initial_values(as_kwargs=False))
        self.bound_up = self._param.get_bounds()[1]
        self.bound_down = self._param.get_bounds()[0]

    def sample(self, init_params, **kwargs_sampler):
        """Minimizes the loss function and returns the best fit."""

        start = time.time()

        if self.sampler == 'emcee':
            samples, logL = self._run_emcee(init_params, **kwargs_sampler)
        elif self.sampler == 'mchmc':
            samples, logL = self._run_mchmc(init_params, **kwargs_sampler)
        else:
            raise NotImplementedError(f'Minimization method {self.sampler} is not supported yet.')

        runtime = time.time() - start
        return samples, logL, runtime

    def _run_emcee(self, initial_values, walker_ratio=5, nsteps=10, sigma_init=1e-4):
        ndim = len(initial_values)
        nwalkers = walker_ratio * ndim
        p0 = initial_values + initial_values * sigma_init * np.random.randn(nwalkers, ndim)

        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.log_likelihood)
        sampler.run_mcmc(p0, nsteps=nsteps, progress=True)

        samples = sampler.get_chain(flat=True)
        logprob = sampler.get_log_prob()

        return samples, logprob

    def _run_mchmc(self, initial_values, num_steps=500, num_chains=10, random_key=None, sigma_init=1e-4):
        mchmcs = MCHMCSampler(self)
        p0 = initial_values + initial_values * sigma_init * np.random.randn(num_chains, self.d)
        samples = mchmcs.sample(num_steps=num_steps, num_chains=num_chains, x_initial=p0, random_key=random_key)
        return samples, None

    def prior_draw(self, key):
        """Uniform prior"""
        return jax.random.uniform(key, shape=(self.d,), minval=self.bound_down, maxval=self.bound_up, dtype='float64')

    def transform(self, x):
        """transform x back to the original parameter. Useful if some parameter are sampled in log space for example"""
        return x



#  
# Copyright (C) 2012-2020 Euclid Science Ground Segment      
#    
# This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General    
# Public License as published by the Free Software Foundation; either version 3.0 of the License, or (at your option)    
# any later version.    
#    
# This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied    
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more    
# details.    
#    
# You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to    
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA    
#

""" 
File: sky_image_plot.py
Created on: Sep 3, 2017
Author: Malte Tewes
This file is a standalone Euclid-agnostic module to visualize and plot sky images with matplotlib.
To visualize Euclid objects, use the wrappers in she_image_checkplot.py
"""

import matplotlib
import numpy as np
from matplotlib import colors
from mpl_toolkits.axes_grid1 import make_axes_locatable


class SkyImage(object):
    """Holds a pixel array
    
    Todo : use properties for z1 z2 !
    """

    def __init__(self, data, z1=None, z2=None):
        """
        
        """

        self.data = data
        self.extent = get_extent(self.data)
        self.z1 = z1
        self.z2 = z2
        self.set_z(z1, z2)

    @property
    def shape(self):  # Just a shortcut
        """The shape (width, height) of the image"""
        return self.data.shape

    def __str__(self):
        return "SkyImage{}[{}:{}]".format(self.shape, self.z1, self.z2)

    def set_z(self, z1=None, z2=None):

        if z1 is None:
            self.z1 = np.min(self.data)
        else:
            self.z1 = z1
        if z2 is None:
            self.z2 = np.max(self.data)
        else:
            self.z2 = z2

        autolist = ["Auto", "auto"]
        if (self.z1 in autolist) or (self.z2 in autolist):
            self.set_auto_z_scale()

    def set_auto_z_scale(self, full_sample_limit=10000, nsig=5.0):
        """Automatic z-scale determination"""

        stata = self.data[:]

        std = stdmad(stata)
        med = np.median(stata)

        nearskypixvals = stata[np.logical_and(stata > med - 2 * std, stata < med + 2 * std)]
        if len(nearskypixvals) > 0:
            skylevel = np.median(nearskypixvals)
        else:
            skylevel = med

        self.z1 = skylevel - nsig * std
        self.z2 = np.max(stata)


def draw_sky_image(ax, si, **kwargs):
    """Use imshow to draw a SkyImage to some axes
    
    """
    # "origin":"lower" as well as the tranpose() within the imshow arguments both combined give the right orientation
    imshow_kwargs = {"aspect": "equal", "origin": "lower", "interpolation": "nearest",
                     "cmap": matplotlib.cm.get_cmap('Greys_r')}
    imshow_kwargs.update(kwargs)
    ax.grid(False)
    ax.set_axis_off()
    im = ax.imshow(si.data.transpose(), extent=si.extent,
                   norm=colors.SymLogNorm(np.abs(si.z1 + 5 * stdmad(si.data[:])), vmin=si.z1 + 5 * stdmad(si.data[:]),
                                          vmax=si.z2), **imshow_kwargs)

    return im


def annotate(ax, cat, x="x", y="y", text="Hello", **kwargs):
    """Annotates the positions (x, y) from a catalog
    
    """

    annotate_kwargs = {"horizontalalignment": "left", "verticalalignment": "top", "color": "red",
                       "xytext": (0, 0), "textcoords": 'offset points'}
    annotate_kwargs.update(**kwargs)

    for row in cat:

        # We skip silently any masked positions
        if getattr(row[x], "mask", False) or getattr(row[y], "mask", False):
            continue

        rowtext = text.format(row=row)
        ax.annotate(rowtext,
                    xy=(row[x], row[y]),
                    **annotate_kwargs
                    )


def get_extent(a):
    """Defines the extent with which to plot an array a (we use the numpy convention)
    
    """
    return (0, a.shape[0], 0, a.shape[1])


def stdmad(a):
    """MAD rescaled to std of normally distributed data"""
    med = np.median(a)
    return 1.4826 * np.median(np.abs(a - med))


def spplot(i, array, fig, title):
    ax = fig.add_subplot(2, 3, i)
    si = SkyImage(np.transpose(array), z1="Auto", z2="Auto")
    im = draw_sky_image(ax, si)
    cat = [{"x": 1.0, "y": array.shape[0] - 1.0, "text": "top left"}]
    annotate(ax, cat, text=title, color="white", fontfamily='sans-serif')

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im, cax=cax, orientation='vertical')



import jax.numpy as jnp
import numpy as np

from starred.utils.parameters import Parameters

__all__ = ['ParametersDeconv']


class ParametersDeconv(Parameters):
    """
    Deconvolution parameters class.

    """
    param_names_analytic = ['a', 'c_x', 'c_y', 'dx', 'dy', 'alpha']
    param_names_background = ['mean', 'h']
    param_names_sersic = ['amp', 'R_sersic', 'n_sersic', 'center_x', 'center_y', 'e1', 'e2']

    def __init__(self, kwargs_init, kwargs_fixed, kwargs_up=None, kwargs_down=None):
        """
        :param kwargs_init: dictionary with information on the initial values of the parameters
        :param kwargs_fixed: dictionary containing the fixed parameters 
        :param kwargs_up: dictionary with information on the upper bounds of the parameters 
        :param kwargs_down: dictionary with information on the lower bounds of the parameters 

        """
        super(ParametersDeconv, self).__init__(kwargs_init, kwargs_fixed, kwargs_up=kwargs_up,
                                               kwargs_down=kwargs_down)

        # we'll guess the number of epochs, etc. based on the provided kwargs
        self.epochs = len(kwargs_init['kwargs_analytic']['dx'])
        self.M = len(kwargs_init['kwargs_analytic']['c_x'])
        self.N_sersic = len(kwargs_init['kwargs_sersic']['amp'])
        self.background_param_number = len(kwargs_init['kwargs_background']['h'])

        self._update_arrays()

    def args2kwargs(self, args):
        """Obtain a dictionary of keyword arguments from positional arguments."""
        i = 0
        kwargs_analytic, i = self._get_params(args, i, 'kwargs_analytic')
        kwargs_background, i = self._get_params(args, i, 'kwargs_background')
        kwargs_sersic, i = self._get_params(args, i, 'kwargs_sersic')
        # wrap-up
        kwargs = {'kwargs_analytic': kwargs_analytic, 'kwargs_background': kwargs_background,
                  'kwargs_sersic': kwargs_sersic}
        return kwargs

    def kwargs2args(self, kwargs):
        """Obtain an array of positional arguments from a dictionary of keyword arguments."""
        args = self._set_params(kwargs, 'kwargs_analytic')
        args += self._set_params(kwargs, 'kwargs_background')
        args += self._set_params(kwargs, 'kwargs_sersic')
        return jnp.array(args)

    def get_param_names_for_model(self, kwargs_key):
        """Returns the names of the parameters according to the key provided."""
        if kwargs_key == 'kwargs_analytic':
            return self.param_names_analytic
        elif kwargs_key == 'kwargs_sersic':
            return self.param_names_sersic
        return self.param_names_background

    def _get_params(self, args, i, kwargs_key):
        """Getting the parameters."""
        kwargs = {}
        kwargs_fixed_k = self._kwargs_fixed[kwargs_key]
        param_names = self.get_param_names_for_model(kwargs_key)
        for name in param_names:
            if not name in kwargs_fixed_k:
                if name == 'a':
                    num_param = self.M * self.epochs
                elif name == 'h':
                    num_param = self.background_param_number
                elif name == 'mean':
                    num_param = self.epochs
                elif name == 'dx' or name == 'dy' or name == 'alpha':
                    num_param = self.epochs
                elif name == 'c_x' or name == 'c_y':
                    num_param = self.M
                elif name in self.param_names_sersic:
                    num_param = self.N_sersic
                else:
                    raise NameError(f"The ParametersDeconv class does not know about this parameter: {name}.")
                kwargs[name] = args[i:i + num_param]
                i += num_param

            else:
                kwargs[name] = kwargs_fixed_k[name]
                free_ind = self._kwargs_free_indices[kwargs_key][name]
                if len(free_ind) > 0:
                    num_param = len(free_ind)
                    kwargs[name] = jnp.array(kwargs[name]).at[free_ind].set(args[i:i + num_param])
                    i += num_param

        return kwargs, i

    def get_all_free_param_names(self, kwargs):
        args = self._param_names(kwargs, 'kwargs_analytic')
        args += self._param_names(kwargs, 'kwargs_background')
        args += self._param_names(kwargs, 'kwargs_sersic')
        return args




from functools import partial

from jax import jit, grad, jacfwd, jacrev, jvp, value_and_grad
import jax.numpy as jnp
import numpy as np

__all__ = ['InferenceBase', 'FisherCovariance']


class InferenceBase(object):
    """Class that defines wraps the loss function, and computes first and second order derivatives.
    
    :param loss_class: Loss instance
    :param param_class: Parameters instance
    """

    def __init__(self, loss_class, param_class):
        self._loss = loss_class
        self._param = param_class

    @property
    def parameters(self):
        """Returns the parameters."""
        return self._param

    # @partial(jit, static_argnums=(0,))
    def loss(self, args):
        """
        Loss function to be minimized.
        :param args: list or array of paramers
        """
        return self._loss(args)

    def log_likelihood(self, args):
        """
        Log Likelihood function to be maximised.
        :param args: list or array of paramers
        """
        return -self._loss(args)

    # @partial(jit, static_argnums=(0,))
    def gradient(self, args):
        """Returns the gradient (first derivative) of the loss function."""
        return grad(self.loss)(args)

    # @partial(jit, static_argnums=(0,))
    def value_and_gradient(self, args):
        """Returns both the value and the gradient (first derivative) of the loss function."""
        return value_and_grad(self.loss)(args)

    # @partial(jit, static_argnums=(0,))
    def hessian(self, args):
        """Returns the Hessian (second derivative) of the loss function."""
        return jacfwd(jacrev(self.loss))(args)

    # @partial(jit, static_argnums=(0,))
    def hessian_vec_prod(self, args, vec):
        """Hessian-vector product."""
        # forward-over-reverse (https://jax.readthedocs.io/en/latest/notebooks/autodiff_cookbook.html#hessian-vector-products-using-both-forward-and-reverse-mode)
        return jvp(grad(self.loss), (args,), (vec,))[1]


# Estimation of the parameter covariance matrix via the Fisher information
# Class copied and modified from Herculens
# Copyright (c) 2022, herculens developers and contributors, Aymeric Galan for this class

class FisherCovariance(object):

    def __init__(self, parameter_class, differentiable_class, diagonal_only=False):
        """
            :param diagonal_only: bool, if you are confident your parameters are not correlated: you can
                                  save space by calculating the diagonal of the hessian only.
        """

        self._param = parameter_class
        self._diff = differentiable_class
        self.diagonal_only = diagonal_only

    @property
    def fisher_matrix(self):
        if not hasattr(self, '_fim'):
            raise ValueError("Call first compute_fisher_information().")
        return self._fim

    @property
    def covariance_matrix(self):
        if not hasattr(self, '_cov'):
            self._cov = self.fisher2covar(self.fisher_matrix, inversion='full')
        return self._cov

    def get_kwargs_sigma(self):
        """
        Return the standard deviation of the marginalized distribution for each parameter. This corresponds to the
        square roots of the diagonal coefficient of the covariance matrix.
        """
        if self.diagonal_only:
            sigma_values = jnp.sqrt(jnp.abs(self.covariance_matrix))
        else:
            sigma_values = jnp.sqrt(jnp.abs(jnp.diag(self.covariance_matrix)))
        kwargs_sigma = self._param.args2kwargs(sigma_values)
        #Set sigma to 0 for the fixed parameters
        for key in kwargs_sigma.keys():
            kwargs_fixed_k = self._param._kwargs_fixed[key]
            for key2 in kwargs_sigma[key].keys():
                if key2 in kwargs_fixed_k:
                    if isinstance(kwargs_sigma[key][key2], (list, np.ndarray, np.generic)):
                        for i in range(len(kwargs_sigma[key][key2])):
                            if i not in self._param._kwargs_free_indices[key][key2]:
                                kwargs_sigma[key][key2][i] = 0.
                    elif isinstance(kwargs_sigma[key][key2], (jnp.ndarray, jnp.generic)):
                        for i in range(len(kwargs_sigma[key][key2])):
                            if i not in self._param._kwargs_free_indices[key][key2]:
                                kwargs_sigma[key][key2] = kwargs_sigma[key][key2].at[i].set(0.)
                    else :
                        kwargs_sigma[key][key2] = 0.

        return kwargs_sigma

    def compute_fisher_information(self, recompute=False):
        """
            :param recompute: bool, redo? default False


        """
        if hasattr(self, '_fim') and not recompute:
            return  # nothing to do
        best_fit_values = self._param.best_fit_values()
        if not self.diagonal_only:
            self._fim = self._diff.hessian(best_fit_values).block_until_ready()
        else:
            hvp = jvp(grad(self._diff.loss), (best_fit_values,), (jnp.ones_like(best_fit_values),))[1]
            self._fim = hvp  # hessian vector product with a vector of ones: diag of hessian.
        self._fim = jnp.array(self._fim)
        if hasattr(self, '_cov'):
            delattr(self, '_cov')

    def draw_samples(self, num_samples=10000, seed=None):
        if seed is not None:
            np.random.seed(seed)
        samples = np.random.multivariate_normal(self._param.best_fit_values(), self.covariance_matrix, size=num_samples)
        return samples

    def fisher2covar(self, fisher_matrix, inversion='full'):
        if self.diagonal_only:
            return 1. / fisher_matrix
        if inversion == 'full':
            return jnp.linalg.inv(fisher_matrix)
        elif inversion == 'diag':
            return 1. / jnp.diag(fisher_matrix)
        else:
            raise ValueError("Only 'full' and 'diag' options are supported for inverting the FIM.")

    @staticmethod
    def split_matrix(matrix, num_before, num_after):
        interior = matrix[num_before:-num_after, num_before:-num_after]

        block_UL = matrix[:num_before, :num_before]
        block_UR = matrix[:num_before, -num_after:]
        block_LR = matrix[-num_after:, -num_after:]
        block_LL = matrix[-num_after:, :num_before]
        exterior = jnp.block([[block_UL, block_UR],
                              [block_LL, block_LR]])

        return interior, exterior



# Adding the STARRED package to sys path so submodules can be called directly

import inspect
import os
import sys

__all__ = ["deconvolution", "plots", "psf", "utils"]

# Needed only if there is no path pointing to the root directory. Mostly for testing purposes.
path_ = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))
sys.path.append(path_)
